# Author: Yiannis Charalambous

from typing import Any, Callable, Dict, Tuple, Type

class SingletonMeta(type):
    """Implements classes as singletons. This is a metaclass, so it should not
    be directly inherited.
    
    If the class you are assigning this as a metaclass of, has a parent that is
    not using this metaclass, you need to use a metaclass factory like makecls
    to automatically resolve conflicts:
    
    ```py
    from noconflict import makecls
    class TestClass(ExistingBase, metaclass=makecls(SingletonMeta)):
        pass
    ```"""
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


# Cache of previously generated metaclasses
metadic: Dict[Tuple[Type[Any], ...], Type[Any]] = {}

def _generatemetaclass(
    bases: Tuple[type, ...],
    metas: Tuple[Type[Any], ...],
    priority: bool
) -> Type[Any]:
    # A metaclass is considered "trivial" if it's the default 'type',
    # or if it's a superclass of any explicitly provided metaclass.
    def trivial(m: Type[Any]) -> bool:
        return sum([issubclass(M, m) for M in metas], int(m is type)) > 0

    # Determine non-trivial metaclasses of the base classes
    metabs: Tuple[Type[Any], ...] = tuple([mb for mb in map(type, bases) if not trivial(mb)])

    # Combine explicit metas and base-class metas according to priority
    # If priority is True, explicit metas come first; else base metas come first
    metabases: Tuple[Type[Any], ...] = (metabs + metas) if not priority else (metas + metabs)

    if metabases in metadic:
        # If a metaclass with this combination of bases was already generated, reuse it
        return metadic[metabases]
    elif not metabases:
        # If no metaclasses are provided, default to Python's 'type'
        meta: Type[Any] = type
    elif len(metabases) == 1:
        # If there's exactly one metaclass, use it directly
        meta = metabases[0]
    else:
        # Multiple metaclasses — generate a new one that inherits from all
        # Create a synthetic name from the names of the base metaclasses
        metaname = "_" + ''.join([m.__name__ for m in metabases])
        meta = makecls()(metaname, metabases, {})  # Recursive class creation with new metaclass

    # Cache and return the new metaclass
    return metadic.setdefault(metabases, meta)

def makecls(
    *metas: Type[Any],
    **options: Any
) -> Callable[[str, Tuple[type, ...], Dict[str, Any]], Type[Any]]:
    """
    Class factory that avoids metaclass conflicts.

    Invocation syntax:
        makecls(M1, M2, ..., priority=True)(name, bases, dict)

    Parameters:
    - metas: Explicit metaclasses to consider when creating the new class.
    - priority: If True, explicit metaclasses are preferred over the bases' metaclasses.

    This function returns a class constructor function. When invoked, it resolves
    all metaclass conflicts among the base classes and the explicitly provided
    metaclasses by dynamically generating a suitable metaclass if necessary.
    """

    # If 'priority' is set, the given metaclasses are considered before base class metaclasses
    priority: bool = options.get('priority', False)
    
    # Return a constructor function that creates the class using the resolved metaclass
    return lambda n, b, d: _generatemetaclass(b, metas, priority)(n, b, d)



# Author: Yiannis Charalambous 2023


from collections import defaultdict
from importlib.machinery import ModuleSpec
from importlib.util import find_spec
import logging
import os
from pathlib import Path

from pydantic_settings import (
    BaseSettings,
    CliPositionalArg,
    NoDecode,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)
from typing import Annotated
from pydantic import (
    AliasChoices,
    BaseModel,
    BeforeValidator,
    DirectoryPath,
    Field,
    FilePath,
    field_validator,
)

from esbmc_ai.singleton import SingletonMeta, makecls
from esbmc_ai.log_handlers import (
    CategoryFileHandler,
    NameFileHandler,
)


def _alias_choice(value: str) -> AliasChoices:
    """Adds aliases to each option that requires a different alias that works
    with all the config setting sources we are using.

    For nested fields, the double underscore notation (e.g., ESBMCAI_VERIFIER__ESBMC__PATH)
    is handled automatically by pydantic_settings via env_nested_delimiter='__'.
    """
    return AliasChoices(
        value,  # exact field name for TOML and direct matching
        value.replace("_", "-"),  # dashed alias for CLI or other uses
        f"ESBMCAI_{value.replace('-', '_').upper()}",  # prefixed env var alias
    )


class AIModelConfig(BaseModel):
    id: str = Field(
        default="openai:gpt-5-nano",
        validation_alias=_alias_choice("ai_model"),
        description="Which AI model to use. Prefix with openai, anthropic, or "
        "ollama then separate with : and enter the model name to use.",
    )

    base_url: str | None = Field(
        default=None,
        exclude=True,
        description="Gets initialized by the config if this model is an Ollama model.",
    )

    @property
    def provider(self) -> str:
        """The provider part of a model string."""
        return self.id.split(":", maxsplit=1)[0]

    @property
    def name(self) -> str:
        """The name part of a model string."""
        return self.id.split(":", maxsplit=1)[1]


class AICustomModelConfig(BaseModel):
    server_type: str
    url: str
    max_tokens: int


class LogConfig(BaseModel):
    output: Path | None = Field(
        default=None,
        description="Save the output logs to a location. Do not add "
        ".log suffix, it will be added automatically.",
    )

    append: bool = Field(
        default=False,
        description="Will append to the logs rather than replace them.",
    )

    by_cat: bool = Field(
        default=False,
        description="Will split the logs by category and write them to"
        " different files. They will have the same base log.output path"
        " but will have an extension to differentiate them.",
    )

    by_name: bool = Field(
        default=False,
        description="Will split the logs by name and write them to"
        " different files. They will have the same base log.output path"
        " but will have an extension to differentiate them.",
    )

    basic: bool = Field(
        default=False,
        description="Enable basic logging mode, will contain no "
        "formatting and also will render --log-by-name (log.by_name) "
        "and --log-by-cat (log.by_cat) useless. Used for debugging "
        "noisy libs.",
    )

    @property
    def logging_handlers(self) -> list[logging.Handler]:
        logging_handlers: list[logging.Handler] = []
        if self.output:
            # Log categories
            if self.by_cat:
                logging_handlers.append(
                    CategoryFileHandler(
                        self.output,
                        append=self.append,
                        skip_uncategorized=True,
                    )
                )
            # Log by name
            if self.by_name:
                logging_handlers.append(
                    NameFileHandler(
                        self.output,
                        append=self.append,
                    )
                )

            # Normal logging
            file_log_handler: logging.Handler = logging.FileHandler(
                str(self.output) + ".log",
                mode="a" if self.append else "w",
            )
            logging_handlers.append(file_log_handler)
        return logging_handlers


class SolutionConfig(BaseModel):
    filenames: CliPositionalArg[list[Path]] = Field(
        default_factory=list,
        description="The filename(s) to pass to the verifier.",
    )

    @field_validator("filenames", mode="after")
    @classmethod
    def on_set_filenames(cls, value: list[str]) -> list[Path]:
        """Validates that filenames are either all file paths or a single directory path."""
        if not value:
            return []

        paths: list[Path] = []
        files: list[str] = []
        dirs: list[str] = []

        for filename in value:
            path = Path(filename).expanduser()

            # Check if path exists
            if not path.exists():
                raise ValueError(f"File or directory not found: '{filename}'")

            # Categorize as file or directory
            if path.is_file():
                files.append(filename)
                paths.append(path)
            elif path.is_dir():
                dirs.append(filename)
                paths.append(path)
            else:
                raise ValueError(
                    f"Path exists but is neither a file nor directory: '{filename}'"
                )

        # Validate: either all files OR single directory
        if dirs and files:
            raise ValueError(
                f"Cannot mix files and directories. Either provide file paths or a single directory.\n"
                f"  Files provided: {', '.join(files)}\n"
                f"  Directories provided: {', '.join(dirs)}"
            )

        if len(dirs) > 1:
            raise ValueError(
                f"Only one directory can be specified. Got {len(dirs)} directories: {', '.join(dirs)}"
            )

        return paths

    include_dirs: list[DirectoryPath] = Field(
        default_factory=list,
        description="Include directories for C files.",
    )

    entry_function: str = Field(
        default="main",
        description="The name of the entry function to repair, defaults to main.",
    )

    output_dir: Path | None = Field(
        default=None,
        description="Set the output directory to save successfully repaired "
        "files in. Leave empty to not use. Specifying the same directory will "
        "overwrite the original file.",
    )

    @field_validator("output_dir", mode="before")
    @classmethod
    def on_set_output_dir(cls, value: Path | None) -> Path | None:
        if value is None:
            return None
        return Path(value).expanduser()

    @field_validator("output_dir", mode="after")
    @classmethod
    def on_after_set_output_dir(cls, value: Path | None) -> Path | None:
        """Creates the directory if it is missing."""
        if value is None:
            return None

        value.mkdir(mode=0o750, parents=True, exist_ok=True)
        return value


class ESBMCConfig(BaseModel):
    """ESBMC-specific configuration.

    Environment variables use double underscore notation for nesting:
    - ESBMCAI_VERIFIER__ESBMC__PATH (handled automatically via env_nested_delimiter)
    """

    path: FilePath | None = Field(
        default=None,
        description="Path to the ESBMC binary.",
    )

    @field_validator("path", mode="before")
    @classmethod
    def on_set_path(cls, value: FilePath | None) -> Path | None:
        if value is None:
            return None
        return Path(value).expanduser()

    params: list[str] = Field(
        default=[
            "--interval-analysis",
            "--goto-unwind",
            "--unlimited-goto-unwind",
            "--k-induction",
            "--state-hashing",
            "--add-symex-value-sets",
            "--k-step",
            "2",
            "--floatbv",
            "--unlimited-k-steps",
            "--context-bound",
            "2",
        ],
        description="Parameters for ESBMC. Can accept as a list or a string.",
    )

    timeout: int | None = Field(
        default=20,
        description="The timeout set for ESBMC.",
    )


class VerifierConfig(BaseModel):
    # The value is checked in AddonLoader.
    name: str = Field(
        default="esbmc",
        description="The verifier to use. Default is ESBMC.",
    )

    enable_cache: bool = Field(
        default=True,
        description="Cache the results of verification in order to save time. "
        "This is not supported by all verifiers.",
    )

    esbmc: ESBMCConfig = Field(
        default_factory=ESBMCConfig,
        description="ESBMC-specific configuration.",
    )


def _parse_ai_model(value: str | dict | AIModelConfig) -> AIModelConfig:
    """Validator function to convert string/dict to AIModelConfig."""
    # If it's already an AIModelConfig, return as-is
    if isinstance(value, AIModelConfig):
        return value
    # If it's a string, create AIModelConfig with validation_alias key
    if isinstance(value, str):
        return AIModelConfig(**{"ai_model": value})
    # If it's a dict, create AIModelConfig from it
    return AIModelConfig(**value)


class Config(BaseSettings, metaclass=makecls(SingletonMeta)):
    """Config loader for ESBMC-AI"""

    config_file: FilePath | None = Field(
        default=None,
        exclude=True,
        description="Path to configuration file (TOML format). Can be set via "
        "ESBMCAI_CONFIG_FILE environment variable.",
    )

    command_name: CliPositionalArg[str] = Field(
        default="help",
        exclude=True,
        description="The (sub-)command to run.",
    )

    addon_modules: list[str] = Field(
        default_factory=list,
        description="The addon modules to load during startup. Additional "
        "modules may be loaded by the specified modules as dependencies.",
    )

    @field_validator("addon_modules", mode="after")
    @classmethod
    def on_set_addon_modules(cls, mods: list[str]) -> list[str]:
        """Validates that a module exists."""
        import sys

        # Temporarily add current directory to sys.path for dev mode addon
        # discovery. This mirrors what AddonLoader does, but we need it here for
        # validation. AddonLoader will add the cwd if needed.
        added_cwd = False
        if "" not in sys.path:
            sys.path.insert(0, "")
            added_cwd = True

        try:
            for m in mods:
                if not isinstance(m, str):
                    raise ValueError("Needs to be a string")
                spec: ModuleSpec | None = find_spec(m)
                if spec is None:
                    raise ValueError(
                        f"Could not find specification for module '{m}'. "
                        "Ensure the module is installed or available in the "
                        "current directory."
                    )
        finally:
            # Clean up: remove the path we added
            if added_cwd and "" in sys.path:
                sys.path.remove("")

        return mods

    verbose_level: int = Field(
        default=0,
        ge=0,
        le=3,
        exclude=True,  # Exclude from Pydantic CLI parsing
        description="Show up to 3 levels of verbose output. Level 1: extra information."
        " Level 2: show failed generations, show ESBMC output. Level 3: "
        "print hidden pushes to the message stack.",
    )

    dev_mode: bool = Field(
        default=False,
        validation_alias=_alias_choice("dev_mode"),
        description="Adds to the python system path the current "
        "directory so addons can be developed.",
    )

    use_json: bool = Field(
        default=False,
        alias="json",
        description="Print the result of the chat command as a JSON output",
    )

    show_horizontal_lines: bool = Field(
        default=True,
        validation_alias=_alias_choice("show_horizontal_lines"),
        description="True to print horizontal lines to segment the output. "
        "Makes it easier to read.",
    )

    horizontal_line_width: int | None = Field(
        default=None,
        validation_alias="horizontal_line_width",
        description="Sets the width of the horizontal lines to draw. "
        "Don't set a value to use the terminal width. Needs to have "
        "show_horizontal_lines set to true.",
    )

    ai_custom: dict[str, AICustomModelConfig] = Field(
        default_factory=defaultdict,
        description=(
            "Dictionary of Ollama AI models configurations. "
            "Each key is a model name (arbitrary string), and each value is a "
            "structure defining the model with the following fields:\n"
            " - server_type: string, must be 'ollama'\n"
            " - url: string specifying the Ollama service URL\n"
            " - max_tokens: integer specifying the token limit"
        ),
    )

    @field_validator("ai_custom", mode="after")
    @classmethod
    def _validate_custom_ai(
        cls, ai_config_list: dict[str, AICustomModelConfig]
    ) -> dict[str, AICustomModelConfig]:
        for name, ai_config in ai_config_list.items():
            # If already an AICustomModelConfig instance, validate its fields
            if isinstance(ai_config, AICustomModelConfig):
                if ai_config.max_tokens <= 0:
                    raise ValueError(
                        f'max_tokens in ai_custom entry "{name}" needs to be greater than 0.'
                    )
            else:
                # If it's a dict, validate the dict structure
                if not isinstance(ai_config, dict):
                    raise ValueError(
                        f"The value of each entry in ai_custom needs to be a dict: {ai_config}"
                    )

                # Max tokens
                if "max_tokens" not in ai_config:
                    raise KeyError(
                        f'max_tokens field not found in "ai_custom" entry "{name}".'
                    )
                elif not isinstance(ai_config["max_tokens"], int):
                    raise TypeError(
                        f'max_tokens in ai_custom entry "{name}" needs to be an int and greater than 0.'
                    )
                elif ai_config["max_tokens"] <= 0:
                    raise ValueError(
                        f'max_tokens in ai_custom entry "{name}" needs to be greater than 0.'
                    )

                # URL
                if "url" not in ai_config:
                    raise KeyError(
                        f'url field not found in "ai_custom" entry "{name}".'
                    )

                # Server type
                if "server_type" not in ai_config:
                    raise KeyError(
                        f"server_type for custom AI '{name}' is invalid, it needs to be a valid string"
                    )

        return ai_config_list

    ai_model: Annotated[AIModelConfig, NoDecode, BeforeValidator(_parse_ai_model)] = (
        Field(
            default_factory=AIModelConfig,
            description="AI Model configuration group. Can be a string like 'openai:gpt-4' or a config object.",
        )
    )

    temp_auto_clean: bool = Field(
        default=True,
        validation_alias=_alias_choice("temp_auto_clean"),
        description="Should the temporary files created be cleared automatically?",
    )

    temp_file_dir: DirectoryPath = Field(
        default=Path("/tmp"),
        validation_alias=_alias_choice("temp_file_dir"),
        description="Sets the directory to store temporary ESBMC-AI files. "
        "Don't supply a value to use the system default.",
    )

    loading_hints: bool = Field(
        default=False,
        validation_alias=_alias_choice("loading_hints"),
        description="Show loading hints when running. Turn off if output "
        "is going to be logged to a file.",
    )

    generate_patches: bool = Field(
        default=False,
        validation_alias=_alias_choice("generate_patches"),
        description="Should the repaired result be returned as a patch "
        "instead of a new file. Generate patch files and place them in "
        "the same folder as the source files.",
    )

    log: LogConfig = Field(
        default_factory=LogConfig,
        description="Logging configuration group.",
    )

    solution: SolutionConfig = Field(
        default_factory=SolutionConfig,
        description="Solution config group.",
    )

    verifier: VerifierConfig = Field(
        default_factory=VerifierConfig,
        description="Verifier config group.",
    )

    llm_requests_max_retries: int = Field(
        default=5,
        ge=1,
        validation_alias="llm_requests.max_retries",
        description="How many times to query the AI service before giving up.",
    )

    llm_requests_timeout: int = Field(
        default=60,
        ge=1,
        validation_alias="llm_requests.timeout",
        description="The timeout for querying the AI service.",
    )

    model_config = SettingsConfigDict(
        env_prefix="ESBMCAI_",
        env_file=".env",
        env_file_encoding="utf-8",
        # Use double underscores for nested config fields in env vars
        # e.g., ESBMCAI_VERIFIER__ESBMC__PATH
        env_nested_delimiter="__",
        # Enables CLI parse support out-of-the-box for CliApp.run integration
        cli_parse_args=True,
        # Allow extra fields for compatibility with pydantic_settings, since
        # .env contains fields for services like OpenAI and Anthropic that isn't
        # mapped to the config directly
        extra="ignore",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type["BaseSettings"],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        # Manually load .env file to get ESBMCAI_CONFIG_FILE before creating sources
        from dotenv import load_dotenv

        load_dotenv(".env", override=False)

        # Get config file path from environment variable
        config_file_path = os.getenv("ESBMCAI_CONFIG_FILE")

        sources: list[PydanticBaseSettingsSource] = [
            init_settings,
            env_settings,
            dotenv_settings,
        ]

        # Add TOML config source if config file is specified
        if config_file_path:
            config_file = Path(config_file_path).expanduser()
            if config_file.exists():
                sources.append(TomlConfigSettingsSource(settings_cls, config_file))

        # Priority order: init/CLI > env > dotenv > TOML > file_secret
        sources.append(file_secret_settings)
        return tuple(sources)

    def __init__(self, **kwargs) -> None:
        # Accept keyword arguments to support pydantic_settings CliApp.run()
        # which passes settings_sources and other configuration parameters.
        # The singleton metaclass intercepts the constructor call and needs
        # to pass these arguments through to BaseSettings.__init__().
        super().__init__(**kwargs)

        # Populate ai_model.base_url from ai_custom if model exists there
        if self.ai_model.id in self.ai_custom:
            self.ai_model.base_url = self.ai_custom[self.ai_model.id].url

    @classmethod
    def set_singleton(cls, config: "Config") -> None:
        """Sets the singleton."""
        getattr(cls, "_instances")[cls] = config



# Author: Yiannis Charalambous

"""This module holds the code for the base source code verifier."""

from pathlib import Path
import re
from typing import Any, override
from hashlib import sha512
import pickle

from platformdirs import user_cache_dir

from esbmc_ai.__about__ import __version__ as esbmc_ai_version
from esbmc_ai.base_component import BaseComponent
from esbmc_ai.log_utils import LogCategories
from esbmc_ai.solution import Solution
from esbmc_ai.verifier_output import VerifierOutput
from esbmc_ai.program_trace import ProgramTrace


class SourceCodeParseError(Exception):
    """Error that means that SolutionGenerator could not parse the source code
    to return the right format."""


class VerifierTimedOutException(Exception):
    """Error that means that ESBMC timed out and so the error could not be
    determined."""


class BaseSourceVerifier(BaseComponent):
    """The base class for creating a source verifier for ESBMC-AI. In order for
    this class to work with ESBMC-AI, the constructor must have default values
    to all arguments because it will be invoked without passing anything.

    Loading from the config will be permitted but it would be preferred if you
    use the base class method `get_config_value`. The fields of the config that
    are going to be loaded need to be declared and returned by the
    `get_config_fields` method. The config loader will automatically preppend
    the verifier_name declared to each key so that there are no clashes in the
    config. So the verifier "esbmc" will have for key "timeout" the following
    field in the config: "esbmc.timeout"."""

    @override
    @classmethod
    def create(cls) -> "BaseComponent":
        obj: BaseComponent = super().create()
        obj._logger = obj.logger.bind(category=LogCategories.VERIFIER)
        return obj

    def __init__(self, verifier_name: str, authors: str) -> None:
        """Verifier name needs to be a valid TOML key."""
        super().__init__()
        self._name = verifier_name
        self._authors = authors

    @property
    def verifier_name(self) -> str:
        """Alias for name"""
        return self.name

    def _cache_name_pack(self, properties: Any) -> Any:
        """Packs additional version properties to the cache name in order to ensure
        it only functions in the current version of ESBMC-AI."""
        return [esbmc_ai_version, properties]

    def _save_cached(self, properties: Any, result: Any) -> None:
        """Saves the verification results to a cached directory to be loaded
        later. Properties are going to be hashed to form the name of the file,
        they should be anything that defines the file."""
        properties = self._cache_name_pack(properties)
        data_dump: bytes = pickle.dumps(obj=properties, protocol=-1)
        file_id: str = str(sha512(data_dump).hexdigest())
        self.logger.info("Saving result to cache")
        self.logger.info(f"Cache ID: {file_id}")

        cache: Path = Path(user_cache_dir("esbmc-ai", "Yiannis Charalambous"))
        cache.mkdir(parents=True, exist_ok=True)
        with open(cache / file_id, "wb") as file:
            pickle.dump(obj=result, file=file, protocol=-1)

    def _load_cached(self, properties: Any) -> Any:
        """Loads the verification results from a cached directory."""
        properties = self._cache_name_pack(properties)
        data_dump: bytes = pickle.dumps(obj=properties, protocol=-1)
        file_id: str = str(sha512(data_dump).hexdigest())
        self.logger.info(f"Searching cache ID: {file_id}")

        cache: Path = Path(user_cache_dir("esbmc-ai", "Yiannis Charalambous"))
        filename: Path = cache / file_id
        if cache.exists() and filename.exists() and filename.is_file():
            with open(filename, "rb") as file:
                data: bytes = pickle.load(file=file)
                self.logger.info("Using cached result")
                return data

        self.logger.info("Cache not found...")
        return None

    def verify_source(
        self,
        *,
        solution: Solution,
    ) -> VerifierOutput:
        """Verifies source_file."""
        _ = solution
        raise NotImplementedError()



# Author: Yiannis Charalambous

"""Contains the base component ABC that is inherited by addons and system
components."""

import inspect
from abc import ABC
import os
from pathlib import Path
import re

from typing import Any, cast, override
from pydantic import FilePath
from pydantic.fields import FieldInfo
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)
from pydantic_settings.sources import InitSettingsSource
import structlog
import tomllib

from esbmc_ai.config import Config


class DictConfigSettingsSource(PydanticBaseSettingsSource):
    """Custom settings source that loads from a pre-loaded dictionary."""

    def __init__(
        self, settings_cls: type[BaseSettings], config_dict: dict[str, Any]
    ) -> None:
        super().__init__(settings_cls)
        self.config_dict = config_dict

    @override
    def get_field_value(
        self, field: FieldInfo, field_name: str
    ) -> tuple[Any, str, bool]:
        """Get field value from the config dictionary."""
        _ = field
        if field_name in self.config_dict:
            return self.config_dict[field_name], field_name, False
        return None, field_name, False

    def __call__(self) -> dict[str, Any]:
        """Return the config dictionary."""
        return self.config_dict


class BaseComponentConfig(BaseSettings):
    """Pydantic BaseSettings preconfigured to be able to load config values.

    Component configs are loaded from the TOML file under the 'addons.<component_name>'
    section by ComponentManager.

    The component name is passed via _component_name kwarg to __init__ and is
    extracted in settings_customise_sources to determine the TOML table header.
    """

    # Used to allow loading from cli and env.
    model_config = SettingsConfigDict(
        env_prefix="ESBMCAI_",
        env_file=".env",
        env_file_encoding="utf-8",
        # Do not parse CLI args in component configs - only the main Config should
        cli_parse_args=False,
        # Ignore extra fields from .env file that don't match the component schema
        extra="ignore",
    )

    def __init__(self, **values: Any) -> None:
        """Used to provide static analyzers type annotations so that we don't get
        errors in ComponentManager."""
        super().__init__(**values)

    @override
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type["BaseSettings"],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        # Get config file path from global Config
        # Note: .env is already loaded by the global Config before component
        # configs are loaded
        config_file_path: FilePath | None = Config().config_file

        sources: list[PydanticBaseSettingsSource] = [
            init_settings,
            env_settings,
            dotenv_settings,
        ]

        # Add TOML config source if config file is specified
        if config_file_path:
            config_file: Path = Path(config_file_path).expanduser()
            if config_file.exists():
                # Cast to InitSettingsSource to access init_kwargs
                init_source: InitSettingsSource = cast(
                    InitSettingsSource, init_settings
                )

                # Extract component name and builtin flag from init_settings if
                # provided
                component_name: str = cast(
                    str, init_source.init_kwargs.get("_component_name")
                )
                builtin: bool = cast(bool, init_source.init_kwargs.get("_builtin"))

                # If a component name is actually given, then the config will be
                # loaded from the file. If none, then nothjing is loaded from the
                # config file, instead the other sources are used only.
                if component_name:
                    # Load TOML file and extract addons.<component_name> section
                    with open(config_file, "rb") as f:
                        config_data: dict = tomllib.load(f)

                    # Get the component-specific config from either <component_name>
                    # or addons.<component_name>
                    component_config: dict = config_data
                    if builtin:
                        component_config = config_data.get(component_name, {})
                    else:
                        component_config = config_data.get("addons", {}).get(
                            component_name, {}
                        )

                    # Add custom dict source with component config
                    if component_config:
                        sources.append(
                            DictConfigSettingsSource(settings_cls, component_config)
                        )

        # Priority order: init > env > dotenv > TOML > file_secret
        sources.append(file_secret_settings)
        return tuple(sources)


class BaseComponent(ABC):
    """The base component class that is inherited by chat commands and verifiers
    and allows them to be loaded by the AddonLoader.

    The model is mapped to the config via the prefix of the name it has. So if
    the name of a BaseComponent is "MyComponent" and we have a Field "my_field"
    then the value from config loaded will be "addons.MyComponent.my_field".
    """

    @classmethod
    def create(cls) -> "BaseComponent":
        """Factory method to instantiate a default version of this class. Used
        by AddonLoader."""
        # Check if __init__ takes only self (no required args)
        sig: inspect.Signature = inspect.signature(cls.__init__)
        params: list[inspect.Parameter] = list(sig.parameters.values())
        # params[0] is always 'self'
        if len(params) > 1 and any(p.default is p.empty for p in params[1:]):
            raise TypeError(
                f"{cls.__name__}.__init__ must take no arguments, override the "
                "create factory method to instantiate the object manually."
            )

        return cls()

    def __init__(self) -> None:
        super().__init__()

        self._global_config: Config
        self._name: str = self.__class__.__name__

        pattern = re.compile(r"[a-zA-Z_]\w*")
        assert pattern.match(
            self._name
        ), f"Invalid toml-friendly verifier name: {self._name}"

        self._authors: str = ""
        self._logger: structlog.stdlib.BoundLogger = structlog.get_logger(
            f"{self._name}"
        )

    @property
    def logger(self) -> structlog.stdlib.BoundLogger:
        """Returns the logger of this class."""
        return self._logger

    @property
    def name(self) -> str:
        """Get the name of this component."""
        return self._name

    @property
    def authors(self) -> str:
        """Get the authors of this component."""
        return self._authors

    @property
    def global_config(self) -> Config:
        """Gets the config for this chat command."""
        return self._global_config

    @global_config.setter
    def global_config(self, value: Config) -> None:
        self._global_config = value

    @property
    def config(self) -> BaseComponentConfig:
        """Gets the component-specific configuration.

        This property provides access to the component's configuration.
        The configuration must be set by the ComponentManager before accessing.

        Returns:
            The component's configuration instance.

        Raises:
            RuntimeError: If configuration has not been set.
        """
        raise NotImplementedError(f"Configuration not set for component {self.name}")

    @config.setter
    def config(self, value: BaseComponentConfig) -> None:
        """Sets the component-specific configuration.

        Args:
            value: The configuration instance for this component.
        """
        _ = value
        raise NotImplementedError(f"Configuration not set for component {self.name}")



# Author: Yiannis Charalambous

"""Contains things related to chat commands."""

from abc import abstractmethod
from typing import Any, override

from esbmc_ai.base_component import BaseComponent
from esbmc_ai.command_result import CommandResult
from esbmc_ai.log_utils import LogCategories


class ChatCommand(BaseComponent):
    """Abstract Base Class for implementing chat commands."""

    @override
    @classmethod
    def create(cls) -> "BaseComponent":
        obj: BaseComponent = super().create()
        obj._logger = obj.logger.bind(category=LogCategories.COMMAND)
        return obj

    def __init__(
        self,
        command_name: str = "",
        authors: str = "",
        help_message: str = "",
    ) -> None:
        super().__init__()
        self._name = command_name
        self._authors = authors

        self.help_message = help_message

    @property
    def command_name(self) -> str:
        """Alias for name"""
        return self.name

    @abstractmethod
    def execute(self) -> CommandResult | None:
        """The main entrypoint of the command. This is abstract and will need to
        be implemented."""
        raise NotImplementedError(f"Command {self.command_name} is not implemented.")



# Author: Yiannis Charalambous

"""This module contains methods to interact and validate prompts."""


def validate_prompt_template_conversation(prompt_template: list[dict]) -> bool:
    """Used to validate if a prompt template conversation is of the correct format
    in the config before loading it."""

    for msg in prompt_template:
        if (
            not isinstance(msg, dict)
            or "content" not in msg
            or "role" not in msg
            or not isinstance(msg["content"], str)
            or not isinstance(msg["role"], str)
        ):
            return False
    return True


def validate_prompt_template(conv: dict[str, list[dict]]) -> bool:
    """Used to check if a prompt template (contains conversation and initial message) is
    of the correct format."""
    if (
        "initial" not in conv
        or not isinstance(conv["initial"], str)
        or "system" not in conv
        or not validate_prompt_template_conversation(conv["system"])
    ):
        return False
    return True



# Author: Yiannis Charalambous

from typing import Any
from uuid import UUID
from langchain_core.messages import AIMessage, BaseMessage
from langchain_core.outputs import ChatGeneration, Generation, LLMResult
from typing_extensions import override
import structlog

from langchain.chat_models import init_chat_model
from langchain_core.language_models import BaseChatModel
from langchain_core.rate_limiters import InMemoryRateLimiter
from langchain_core.callbacks import BaseCallbackHandler


from esbmc_ai.config import Config
from esbmc_ai.log_utils import LogCategories


class LoggingCallbackHandler(BaseCallbackHandler):
    """Invoke callback handler is used to print debug messages to the LLM."""

    def __init__(self, ai_model: str) -> None:
        super().__init__()
        self.logger: structlog.stdlib.BoundLogger = structlog.get_logger().bind(
            category=LogCategories.CHAT,
            prefix_name=ai_model,
        )
        # Track last printed message index per group: {group_idx: last_msg_idx}
        self._last_printed_idx: dict[int, int] = {}

    @staticmethod
    def _get_msg_formatted(
        group_idx: int, msg_idx: int, msg: BaseMessage | Generation
    ) -> str:
        base: str = f"MSG {group_idx}-{msg_idx} {msg.type.capitalize()}:"
        if isinstance(msg, AIMessage) and msg.tool_calls:
            return f"{base} <message-body>{msg.text}</message-body>\nTool Call: {msg.tool_calls}"
        return f"{base} <message-body>{msg.text}</message-body>"

    # @override
    # def on_llm_start(
    #     self,
    #     serialized: dict[str, Any],
    #     prompts: list[str],
    #     *,
    #     run_id: UUID,
    #     parent_run_id: UUID | None = None,
    #     tags: list[str] | None = None,
    #     metadata: dict[str, Any] | None = None,
    #     **kwargs: Any,
    # ) -> Any:
    #     """Run when LLM starts running.

    #     .. ATTENTION::
    #         This method is called for non-chat models (regular LLMs). If you're
    #         implementing a handler for a chat model, you should use
    #         ``on_chat_model_start`` instead.

    #     Args:
    #         serialized (dict[str, Any]): The serialized LLM.
    #         prompts (list[str]): The prompts.
    #         run_id (UUID): The run ID. This is the ID of the current run.
    #         parent_run_id (UUID): The parent run ID. This is the ID of the parent run.
    #         tags (Optional[list[str]]): The tags.
    #         metadata (Optional[dict[str, Any]]): The metadata.
    #         kwargs (Any): Additional keyword arguments.
    #     """
    #     _ = run_id, parent_run_id, tags, metadata, kwargs
    #     self.logger.debug("Invoke LLM")

    @override
    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        _ = run_id, parent_run_id, kwargs
        self.logger.debug("=" * 80)
        self.logger.debug("LLM Response")
        for idx, msg_group in enumerate(response.generations):
            self.logger.debug("=" * 80)
            for msg_idx, msg in enumerate(msg_group):
                if isinstance(msg, ChatGeneration):
                    self.logger.debug(
                        self._get_msg_formatted(idx, msg_idx, msg.message)
                    )
                else:
                    self.logger.debug(self._get_msg_formatted(idx, msg_idx, msg))

                if msg_idx < len(msg_group) - 1:
                    self.logger.debug("-" * 80)
        self.logger.debug("=" * 80)

    @override
    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        _ = run_id, parent_run_id, kwargs
        self.logger.debug("=" * 80)
        self.logger.debug("LLM Error")
        self.logger.debug(str(error))
        self.logger.debug("=" * 80)

    @override
    def on_chat_model_start(
        self,
        serialized: dict[str, Any],
        messages: list[list[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: list[str] | None = None,
        metadata: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when a chat model starts running.

        **ATTENTION**: This method is called for chat models. If you're implementing
        a handler for a non-chat model, you should use ``on_llm_start`` instead.

        Args:
            serialized (dict[str, Any]): The serialized chat model.
            messages (list[list[BaseMessage]]): The messages.
            run_id (UUID): The run ID. This is the ID of the current run.
            parent_run_id (UUID): The parent run ID. This is the ID of the parent run.
            tags (Optional[list[str]]): The tags.
            metadata (Optional[dict[str, Any]]): The metadata.
            kwargs (Any): Additional keyword arguments.
        """
        _ = run_id, parent_run_id, tags, metadata, kwargs
        self.logger.debug("=" * 80)
        self.logger.debug("Invoke Chat Model LLM")
        for group_idx, msg_group in enumerate(messages):
            last_printed = self._last_printed_idx.get(group_idx, -1)
            self.logger.debug("=" * 80)
            for msg_idx, msg in enumerate(
                msg_group[last_printed + 1 :], start=last_printed + 1
            ):
                self.logger.debug(self._get_msg_formatted(group_idx, msg_idx, msg))
                if msg_idx < len(msg_group) - 1:
                    self.logger.debug("-" * 80)
            self._last_printed_idx[group_idx] = len(msg_group) - 1


class AIModel:
    """Loading utils for models."""

    @classmethod
    def get_model(
        cls,
        *,
        model: str,
        provider: str | None = None,
        temperature: float | None = None,
        url: str | None = None,
    ) -> BaseChatModel:
        handler: BaseCallbackHandler = LoggingCallbackHandler(ai_model=model)

        chat_model: BaseChatModel = init_chat_model(
            model=model,
            model_provider=provider,
            temperature=temperature,
            max_tokens=None,  # Use all remaining tokens
            base_url=url,
            timeout=Config().llm_requests_timeout,
            max_retries=Config().llm_requests_max_retries,
            rate_limiter=InMemoryRateLimiter(
                requests_per_second=10,
                check_every_n_seconds=0.1,
                max_bucket_size=100,
            ),
            callbacks=[handler],
        )

        return chat_model



# Author: Yiannis Charalambous 2023

"""Contains code for automatically repairing code using ESBMC."""

from langchain_core.prompts import PromptTemplate
from langchain_core.messages import BaseMessage
from langchain_core.language_models import BaseChatModel

from esbmc_ai.solution import Solution
from esbmc_ai.chats.template_key_provider import (
    OracleTemplateKeyProvider,
    TemplateKeyProvider,
)
from esbmc_ai.verifiers.esbmc import ESBMCOutput
from esbmc_ai.chats import KeyTemplateRenderer


class SolutionGenerator:
    """SolutionGenerator is a simple conversation-based automated program repair
    class. It maintains a conversation with the LLM, starting with a system message
    (provided at initialization) and then adding repair attempt messages via
    generate_solution calls."""

    def __init__(
        self,
        ai_model: BaseChatModel,
        esbmc_output_type: str = "full",
        system_message: list[BaseMessage] | None = None,
    ) -> None:
        """Initializes the solution generator."""
        super().__init__()

        self.ai_model: BaseChatModel = ai_model
        self.template_key_provider: TemplateKeyProvider = OracleTemplateKeyProvider()
        self.messages: list[BaseMessage] = system_message or []

        self.esbmc_output_type: str = esbmc_output_type

        self.invokations: int = 0

    @staticmethod
    def extract_code_from_solution(solution: str) -> str:
        """Strip the source code of any leftover text as sometimes the AI model
        will generate text and formatting despite being told not to."""
        try:
            code_start: int = solution.index("```") + 3
            assert code_start != -1

            # Remove up until the new line, because usually there's a language
            # specification after the 3 ticks ```c...
            code_start = solution.index("\n", code_start) + 1
            assert code_start != -1

            code_end: int = solution[::-1].index("```")
            assert code_start != -1

            # -4 = 3 backticks and also the \n before the backticks.
            code_end: int = len(solution) - 4 - code_end
            # +1 because of edge cases as in test_get_code_from_solution
            assert code_start <= code_end + 1

            solution = solution[code_start:code_end]
        except (ValueError, AssertionError):
            pass
        return solution

    def generate_solution(
        self,
        initial_message_prompt: PromptTemplate,
        solution: Solution,
        verifier_output: ESBMCOutput,
    ) -> str:
        """Prompts the LLM to repair the source code using the verifier output.
        Returns the extracted code from the LLM's response."""

        # Add the initial message for this repair attempt
        # Pass the template string to KeyTemplateRenderer which will handle formatting
        key_template_renderer: KeyTemplateRenderer = KeyTemplateRenderer(
            messages=[("human", initial_message_prompt.template)],
            key_provider=self.template_key_provider,
        )

        # Format with the solution and oracle output
        formatted_messages = key_template_renderer.format_messages(
            solution=solution,
            oracle_output=verifier_output,
        )
        self.messages.extend(formatted_messages)

        self.invokations += 1

        # Generate the solution
        response: BaseMessage = self.ai_model.invoke(self.messages)

        # Add AI response to message history for conversation context
        self.messages.append(response)

        repaired_code = SolutionGenerator.extract_code_from_solution(
            str(response.content)
        )

        return repaired_code



# Author: Yiannis Charalambous

from enum import Enum
from pathlib import Path
from typing import override
import logging
import re

from esbmc_ai.log_categories import LogCategories

_ansi_escape = re.compile(r"\x1b\[[0-9;]*m")


def _strip_ansi_escape_processor(record: logging.LogRecord) -> bool | logging.LogRecord:
    """
    Remove ANSI escape sequences from all string values in the LogRecord.
    """
    for attr, value in list(record.__dict__.items()):
        if isinstance(value, str):
            setattr(record, attr, _ansi_escape.sub("", value))

    args = record.args
    if isinstance(args, tuple):
        record.args = tuple(
            _ansi_escape.sub("", v) if isinstance(v, str) else v for v in args
        )
    elif isinstance(args, dict):
        record.args = {
            k: (_ansi_escape.sub("", v) if isinstance(v, str) else v)
            for k, v in args.items()
        }

    return record


class CategoryFileHandler(logging.Handler):
    """Logger that will save by category."""

    def __init__(
        self,
        base_path: Path,
        append: bool = False,
        skip_uncategorized: bool = False,
    ) -> None:
        super().__init__()
        self.base_path = base_path
        self.append = append
        self.skip_uncategorized = skip_uncategorized
        self.handlers: dict[str, logging.FileHandler] = {}
        self.addFilter(_strip_ansi_escape_processor)

    def emit(self, record: logging.LogRecord) -> None:
        # Grab the category (because of wrap_for_formatter)
        # Try attribute first (for stdlib logging)
        category: str | Enum | None = getattr(record, "category", None)

        # If not present, try to get it from record.msg (structlog dict)
        if category is None and isinstance(record.msg, dict):
            category = record.msg.get("category", None)

        # Convert to string
        if isinstance(category, Enum):
            category = category.value

        assert isinstance(category, str), f"Category is not string: {category}"

        # Skip uncategorized if desired
        if (
            not category or category == LogCategories.NONE.value
        ) and self.skip_uncategorized:
            return

        # None category is a catch all
        if not category:
            category = LogCategories.NONE.value

        # Write ALL category
        if category == LogCategories.ALL.value:
            for h in self.handlers.values():
                h.emit(record)
            return

        # Lazily build the per‐category FileHandler
        if category not in self.handlers:
            fn: Path = Path(f"{self.base_path}-{category}.log")
            fh: logging.FileHandler = logging.FileHandler(
                fn, mode="a" if self.append else "w"
            )
            fh.setFormatter(self.formatter)
            self.handlers[category] = fh

        # Delegate to the per‐category handler
        self.handlers[category].emit(record)


class NameFileHandler(logging.Handler):
    """Logging file handler that will write by logger name."""

    def __init__(
        self, base_path: Path, append: bool = False, skip_unnamed: bool = False
    ) -> None:
        super().__init__()
        self.base_path: Path = base_path
        self.append: bool = append
        self.handlers: dict[str, logging.FileHandler] = {}
        self.skip_unnamed: bool = skip_unnamed
        self.addFilter(_strip_ansi_escape_processor)

    @override
    def emit(self, record: logging.LogRecord) -> None:
        logger_name: str = record.name
        # Write to file (and also to stdout, if desired)
        if logger_name not in self.handlers:
            handler = logging.FileHandler(
                f"{self.base_path}-{logger_name}.log",
                mode="a" if self.append else "w",
            )
            handler.setFormatter(self.formatter)
            self.handlers[logger_name] = handler
        self.handlers[logger_name].emit(record)



# Author: Yiannis Charalambous

"""Contains the help command."""

from typing import Any
from typing_extensions import override

from esbmc_ai.chat_command import ChatCommand
from esbmc_ai.component_manager import ComponentManager


class HelpCommand(ChatCommand):
    """Command that prints helpful information about other commands. Including
    addon commands."""

    commands: list[ChatCommand] = []

    def __init__(self) -> None:
        super().__init__(
            command_name="help",
            help_message="Print this help message.",
        )

    @override
    def execute(self) -> Any:
        print("Commands:")
        for command in ComponentManager().builtin_commands.values():
            print(f"* {command.command_name}: {command.help_message}")
            if command.authors:
                print(f"\tAuthors: {command.authors}")

        if ComponentManager().addon_commands:
            print("\nAddon Commands:")
        else:
            self.logger.info("No addon commands to show...")
        for command in ComponentManager().addon_commands.values():
            print(f"* {command.command_name}: {command.help_message}")
            if command.authors:
                print(f"\tAuthors: {command.authors}")

        print("\nLicense:")
        print("* ESBMC-AI is dual-licensed (AGPL-3.0 / Commercial)")
        print("* Run 'esbmc-ai license' for full details")



"""ESBMC-AI is an AI augmentation layer over ESBMC, the Efficient
Software Bounded Model Checker. With the power of LLMs, it adds
features such as automatic code fixing and more."""

from esbmc_ai.__about__ import __version__, __author__

from esbmc_ai.config import Config
from esbmc_ai.chat_command import ChatCommand
from esbmc_ai.base_component import BaseComponent, BaseComponentConfig
from esbmc_ai.verifiers import BaseSourceVerifier
from esbmc_ai.ai_models import AIModel
from esbmc_ai.log_categories import LogCategories
from esbmc_ai.chats.key_template_renderer import KeyTemplateRenderer
from esbmc_ai.chats.template_key_provider import (
    TemplateKeyProvider,
    GenericTemplateKeyProvider,
    OracleTemplateKeyProvider,
)

__all__ = [
    "Config",
    "BaseComponent",
    "BaseComponentConfig",
    "ChatCommand",
    "BaseSourceVerifier",
    "AIModel",
    "LogCategories",
    "KeyTemplateRenderer",
    "TemplateKeyProvider",
    "GenericTemplateKeyProvider",
    "OracleTemplateKeyProvider",
]



# Author: Yiannis Charalambous

"""Template renderer that integrates TemplateKeyProvider with ChatPromptTemplate."""

from typing import Any, Sequence, override
from jinja2.sandbox import SandboxedEnvironment
from langchain_core.messages import BaseMessage
from langchain_core.prompt_values import ChatPromptValue
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.prompts.chat import MessageLikeRepresentation
from langchain_core.prompts.string import PromptTemplateFormat, DEFAULT_FORMATTER_MAPPING
from pydantic import PrivateAttr, BaseModel
from esbmc_ai.chats.template_key_provider import TemplateKeyProvider
from .template_funcs import get_func_mapping


class _PermissiveSandboxedEnvironment(SandboxedEnvironment):
    """Custom Jinja2 sandbox that allows attribute access on trusted objects.

    This environment allows attribute access on Pydantic models and other
    trusted objects while still providing basic sandboxing protections.
    """

    def is_safe_attribute(self, obj: Any, attr: str, value: Any) -> bool:
        """Allow attribute access on Pydantic models and common types."""
        # Allow attribute access on Pydantic models
        if isinstance(obj, BaseModel):
            return True
        # Allow attribute access on dicts, lists, and other safe types
        if isinstance(obj, (dict, list, tuple, str, int, float, bool, type(None))):
            return True
        # Default to parent's behavior for other types
        return super().is_safe_attribute(obj, attr, value)


def _permissive_jinja2_formatter(template: str, **kwargs: Any) -> str:
    """Format a template using a permissive Jinja2 sandbox.

    Unlike LangChain's default restricted sandbox, this allows attribute
    access on trusted objects like Pydantic models.
    """
    env = _PermissiveSandboxedEnvironment()
    return env.from_string(template).render(**kwargs)


# Override LangChain's default jinja2 formatter with our permissive one
# This allows nested attribute access on Pydantic models in templates
DEFAULT_FORMATTER_MAPPING["jinja2"] = _permissive_jinja2_formatter


class KeyTemplateRenderer(ChatPromptTemplate):
    """Derives ChatPromptTemplate and automatically provides template keys via
    TemplateKeyProvider. This is to force standarization across types and keys."""

    _key_provider: TemplateKeyProvider = PrivateAttr()

    def __init__(
        self,
        messages: Sequence[MessageLikeRepresentation],
        key_provider: TemplateKeyProvider,
        *,
        template_format: PromptTemplateFormat = "jinja2",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            messages=messages,
            template_format=template_format,
            **kwargs,
        )
        self._key_provider = key_provider

    @override
    def format_prompt(self, **kwargs: Any) -> ChatPromptValue:
        auto_keys = self._key_provider.get_template_keys(**kwargs)
        kwargs = {**auto_keys, **kwargs} | get_func_mapping()
        return super().format_prompt(**kwargs)

    @override
    def format_messages(self, **kwargs: Any) -> list[BaseMessage]:
        auto_keys = self._key_provider.get_template_keys(**kwargs)
        kwargs = {**auto_keys, **kwargs} | get_func_mapping()
        return super().format_messages(**kwargs)

    @override
    def format(self, **kwargs: Any) -> str:
        auto_keys = self._key_provider.get_template_keys(**kwargs)
        kwargs = {**auto_keys, **kwargs} | get_func_mapping()
        return super().format(**kwargs)



#!/usr/bin/env python3

# Author: Yiannis Charalambous 2023

import logging
import sys

import argparse

from pydantic_core import ErrorDetails
from pydantic_settings import CliApp, CliSettingsSource
from pydantic_core._pydantic_core import ValidationError
from structlog import get_logger
from structlog.stdlib import BoundLogger

from esbmc_ai import Config, ChatCommand, __author__, __version__
from esbmc_ai.addon_loader import AddonLoader
from esbmc_ai.command_result import CommandResult
from esbmc_ai.log_utils import LogCategories, get_log_level, init_logging
from esbmc_ai.verifiers.base_source_verifier import BaseSourceVerifier
from esbmc_ai.verifiers.esbmc import ESBMC
from esbmc_ai.component_manager import ComponentManager
import esbmc_ai.commands


def _init_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-h",
        "--help",
        action="store_true",
        help="Show this help message and exit.",
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
        help="Show version information.",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help=(
            "Show up to 3 levels of verbose output. Level 1: extra information."
            " Level 2: show failed generations, show ESBMC output. Level 3: "
            "print hidden pushes to the message stack."
        ),
    )


def _load_config(
    parser: argparse.ArgumentParser,
) -> Config:
    # Check if help is requested - if so, skip verbose level extraction
    # and let CliApp.run handle the full help display with all Pydantic fields
    help_requested = "--help" in sys.argv or "-h" in sys.argv

    if not help_requested:
        # Parse args to get verbose level before Pydantic CLI parsing
        args, _ = parser.parse_known_args()
        verbose_level: int = min(args.verbose, 3) if hasattr(args, "verbose") else 0
    else:
        # No need to parse verbosity.
        verbose_level: int = 0

    # Create custom CLI settings source using our argparse parser
    # This allows us to use custom arguments like -v/--verbose with action='count'
    cli_settings: CliSettingsSource = CliSettingsSource(
        Config, cli_parse_args=True, root_parser=parser, cli_implicit_flags=True
    )

    # If help is requested and validation fails, catch error and show help anyway
    try:
        # Use CliApp.run with custom CLI settings source
        # settings_customise_sources will handle TOML/env/dotenv loading process
        config: Config = CliApp.run(Config, cli_settings_source=cli_settings)
    except ValidationError as e:
        if help_requested:
            # When help is requested, show help even if config validation fails
            # CliSettingsSource has already added all Pydantic fields to the parser
            parser.print_help()
            sys.exit(0)
        else:
            # Re-raise the exception if help was not requested
            print(f"ESBMC-AI: validation error while loading {e.title}\n")
            errs: list[ErrorDetails] = e.errors(
                include_context=True,
                include_input=True,
            )
            for idx, err in enumerate(errs):
                print(
                    f"* Error {idx}: "
                    f'{err["type"]}: {err["loc"]} cannot accept '
                    f'"{err["input"]}" (type {type(err["input"]).__name__}) '
                    f'because: {err["msg"]}'
                )
            print("\nShowing traceback...")
            print("=" * 80)
            raise

    # If help was requested and config loaded successfully, show help and exit
    # This handles the case where --help is passed but config validation succeeds
    if help_requested:
        parser.print_help()
        sys.exit(0)

    # Set argparse config values - These fields have exclude=True
    config.verbose_level = verbose_level

    return config


def _init_builtin_components() -> None:
    """Initializes the builtin verifiers and commands."""
    component_manager = ComponentManager()

    # Built-in verifiers
    esbmc = ESBMC.create()
    assert isinstance(esbmc, BaseSourceVerifier)
    component_manager.add_verifier(esbmc)
    # Load component-specific configuration
    component_manager.load_component_config(esbmc, builtin=True)

    # Init built-in commands - Loads everything in the esbmc_ai.commands module.
    commands: list[ChatCommand] = []
    for cmd_classname in getattr(esbmc_ai.commands, "__all__"):
        cmd_type: type = getattr(esbmc_ai.commands, cmd_classname)
        assert issubclass(cmd_type, ChatCommand), f"{cmd_type} is not a ChatCommand"
        cmd: object = cmd_type.create()
        assert isinstance(cmd, ChatCommand)
        cmd.global_config = Config()
        # Load component-specific configuration
        component_manager.load_component_config(cmd, builtin=True)
        commands.append(cmd)

    component_manager.set_builtin_commands(commands)


def _init_logging() -> None:
    # Add logging handlers with config options
    config = Config()
    logging_handlers: list[logging.Handler] = config.log.logging_handlers

    # Reinit logging
    init_logging(
        level=get_log_level(config.verbose_level),
        file_handlers=logging_handlers,
        init_basic=config.log.basic,
    )


def main() -> None:
    """Entry point function"""
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="ESBMC-AI",
        description="""Automated Program Repair platform. To view help on subcommands, run with the subcommand "help".

Configuration Precedence (highest to lowest):
  * CLI args > Environment variables > .env file > TOML config > Defaults
  * NOTE: Nested config values can be set via environment variables using double underscores
    (e.g., ESBMCAI_VERIFIER__ESBMC__PATH for verifier.esbmc.path)""",
        epilog=f"Made by {__author__}",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        # Disable argparse help to let Pydantic handle full help display
        add_help=False,
    )

    _init_args(parser=parser)

    config: Config
    config = _load_config(parser=parser)
    # Set the config singleton
    Config.set_singleton(config)
    config: Config = Config()

    print(f"ESBMC-AI v{__version__}")
    print(f"Made by {__author__}")
    print()

    _init_logging()

    logger: BoundLogger = get_logger().bind(category=LogCategories.SYSTEM)
    logger.debug("Global config loaded successfully")
    logger.debug("Initialized logging")

    _init_builtin_components()
    logger.debug("Builtin components loaded successfully")

    if config.dev_mode:
        logger.warn("Development Mode Activated")

    # Load addons
    addon_loader: AddonLoader = AddonLoader(config)
    logger.debug("Addon components loaded successfully")
    logger.info("Configuration loaded successfully")

    # Bind addons to component loader.
    cm = ComponentManager()
    for command in addon_loader.chat_command_addons.values():
        cm.add_command(command, builtin=False)

    for verifier in addon_loader.verifier_addons.values():
        cm.add_verifier(verifier, builtin=False)

    cm.set_verifier_by_name(config.verifier.name)

    # Run the command
    command_name = config.command_name
    command_names: list[str] = cm.command_names
    if command_name in command_names:
        logger.info(f"Running Command: {command_name}\n")
        command: ChatCommand = cm.commands[command_name]
        result: CommandResult | None = command.execute()
        if result:
            if config.use_json:
                print(vars(result))
            else:
                print(result)

        sys.exit(0)
    else:
        logger.error(
            f"Error: Unknown command: {command_name}. Choose from: "
            + ", ".join(command_names)
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



# Author: Yiannis Charalambous

"""Contains the exit chat command."""

import sys
from typing import Any, Optional
from typing_extensions import override

from esbmc_ai.chat_command import ChatCommand


class ExitCommand(ChatCommand):
    """Used to exit user chat mode gracefully."""

    def __init__(self) -> None:
        super().__init__(
            command_name="exit",
            help_message="Exit the program.",
        )

    @override
    def execute(self) -> None:
        print("exiting...")
        sys.exit(0)



# Author: Yiannis Charalambous

from enum import Enum
import os
from pathlib import Path
from typing import Any
from pydantic import Field, field_validator
from typing_extensions import override
from langchain_core.prompts import PromptTemplate
from langchain_core.messages import BaseMessage, SystemMessage, HumanMessage, AIMessage

from esbmc_ai.base_component import BaseComponentConfig
from esbmc_ai.component_manager import ComponentManager
from esbmc_ai.solution import Solution, SourceFile
from esbmc_ai.ai_models import AIModel
from esbmc_ai.chats.solution_generator import SolutionGenerator
from esbmc_ai.command_result import CommandResult
from esbmc_ai.verifier_output import VerifierOutput
from esbmc_ai.chat_command import ChatCommand
from esbmc_ai.loading_widget import BaseLoadingWidget, LoadingWidget
from esbmc_ai.verifiers.esbmc import ESBMC, ESBMCOutput


class FixCodeCommandResult(CommandResult):
    """Returned by the FixCodeCommand"""

    def __init__(
        self,
        successful: bool,
        attempts: int,
        repaired_source: str | None = None,
    ) -> None:
        super().__init__()
        self._successful: bool = successful
        self.attempts: int = attempts
        self.repaired_source: str | None = repaired_source

    @property
    @override
    def successful(self) -> bool:
        return self._successful

    @override
    def __str__(self) -> str:
        if self._successful and self.repaired_source is not None:
            return self.repaired_source

        return "Failed all attempts..."


class FixCodeCommandConfig(BaseComponentConfig):
    class VerifierOutputType(str, Enum):
        full = "full"
        ce = "ce"
        vp = "vp"

    verifier_output_type: str = Field(
        default=VerifierOutputType.full,
        description="The type of output from ESBMC in the fix code command.",
    )

    temperature: float = Field(
        default=0,
        description="The temperature of the LLM for the fix code command.",
    )

    max_attempts: int = Field(
        default=5,
        description="Fix code command max attempts.",
    )

    initial: str = Field(
        default="ESBMC found an error in the code:\n\nError Type: {{oracle_output.error_type}}\nError Message: {{oracle_output.error_message}}\nError Location: {{oracle_output.error_file}}:{{oracle_output.error_line}}\n\nStack Trace:\n{{oracle_output.primary_issue.stack_trace_formatted}}\n\n{% if is_verifier_issue(oracle_output.primary_issue) and oracle_output.primary_issue.counterexample | length > 0 %}Counterexample:\n{{oracle_output.primary_issue.counterexample_formatted}}\n\n{% endif %}The source code is:\n\n```c\n{{solution.files[0].content}}\n```\n\nUsing the error information above, show the fixed text.",
        description="Initial prompt for the first repair attempt. Uses structured oracle output fields.",
    )

    retry_prompt: str = Field(
        default="The previous attempt failed. ESBMC found an error:\n\nError Type: {{oracle_output.error_type}}\nError Message: {{oracle_output.error_message}}\nError Location: {{oracle_output.error_file}}:{{oracle_output.error_line}}\n\nStack Trace:\n{{oracle_output.primary_issue.stack_trace_formatted}}\n\n{% if is_verifier_issue(oracle_output.primary_issue) and oracle_output.primary_issue.counterexample | length > 0 %}Counterexample:\n{{oracle_output.primary_issue.counterexample_formatted}}\n\n{% endif %}The source code is:\n\n```c\n{{solution.files[0].content}}\n```\n\nPlease review the conversation history to see what was tried before. Using the error information above and learning from previous failed attempts, show the fixed text.",
        description="Prompt used for retry attempts after the initial attempt fails. Uses structured oracle output fields and can reference conversation history.",
    )

    system: list[dict[str, str]] = [
        {
            "role": "system",
            "content": "From now on, act as an Automated Code Repair Tool that repairs C code. You will be shown C code along with structured error information from ESBMC verification. Pay close attention to the error type, error message, and error location provided. Use this information to identify and fix the bug. Provide the repaired C code as output, as would an Automated Code Repair Tool. Aside from the corrected source code, do not output any other text.",
        }
    ]

    @field_validator("verifier_output_type", mode="after")
    @classmethod
    def validate_verifier_output_type(cls, v: str) -> str:
        if v not in ["full", "vp", "ce"]:
            raise ValueError("verifier_output_type must be 'full', 'vp', or 'ce'")
        return v


class FixCodeCommand(ChatCommand):
    """Command for automatically fixing code using a verifier."""

    def __init__(self) -> None:
        super().__init__(
            command_name="fix-code",
            help_message="Generates a solution for this code, and reevaluates it with ESBMC.",
        )
        # Set default config instance
        self._config: FixCodeCommandConfig = FixCodeCommandConfig()
        self.original_source_file: SourceFile
        self.anim: BaseLoadingWidget

    @classmethod
    def _get_config_class(cls) -> type[BaseComponentConfig]:
        """Return the config class for this component."""
        return FixCodeCommandConfig

    @property
    @override
    def config(self) -> BaseComponentConfig:
        return self._config

    @config.setter
    def config(self, value: BaseComponentConfig) -> None:
        assert isinstance(value, FixCodeCommandConfig)
        self._config = value

    @override
    def execute(self) -> FixCodeCommandResult:
        self.logger.warn(
            "Fix Code Command is a demonstration of the ESBMC-AI "
            + "API capabilities. Do not use for advanced repair tasks.\n"
        )
        # Handle kwargs
        source_file: SourceFile = SourceFile.load(
            self.global_config.solution.filenames[0]
        )
        self.original_source_file = SourceFile(
            file_path=source_file.file_path, content=source_file.content
        )
        self.anim = (
            LoadingWidget() if self.global_config.loading_hints else BaseLoadingWidget()
        )
        # End of handle kwargs

        solution: Solution = Solution([])
        solution.add_source_file(source_file)

        self._logger.info(f"FixCodeConfig: {self._config}")

        verifier: Any = ComponentManager().get_verifier("esbmc")
        assert isinstance(verifier, ESBMC)
        verifier_output: VerifierOutput = verifier.verify_source(solution=solution)
        assert isinstance(verifier_output, ESBMCOutput)

        if verifier_output.successful:
            self.logger.info("File verified successfully")
            returned_source: str
            if self.global_config.generate_patches:
                returned_source = source_file.get_diff(source_file)
            else:
                returned_source = source_file.content
            return FixCodeCommandResult(True, 0, returned_source)

        # Create the AI model with the specified parameters
        ai_model = AIModel.get_model(
            model=self.global_config.ai_model.id,
            temperature=self._config.temperature,
            url=self.global_config.ai_model.base_url,
        )

        # Convert system messages from dict format to BaseMessage objects
        system_messages: list[BaseMessage] = []
        for msg in self._config.system:
            role = msg.get("role", "system")
            content = msg.get("content", "")
            if role == "system":
                system_messages.append(SystemMessage(content=content))
            elif role == "human":
                system_messages.append(HumanMessage(content=content))
            elif role == "assistant" or role == "ai":
                system_messages.append(AIMessage(content=content))

        # Create the initial and retry message templates
        initial_prompt = PromptTemplate.from_template(self._config.initial)
        retry_prompt = PromptTemplate.from_template(self._config.retry_prompt)

        solution_generator: SolutionGenerator = SolutionGenerator(
            ai_model=ai_model,
            system_message=system_messages,
            esbmc_output_type=self._config.verifier_output_type,
        )

        print()

        for attempt in range(1, self._config.max_attempts + 1):
            # Use initial prompt for first attempt, retry prompt for subsequent attempts
            prompt = initial_prompt if attempt == 1 else retry_prompt

            result: FixCodeCommandResult | None
            result, verifier_output = self._attempt_repair(
                attempt=attempt,
                solution_generator=solution_generator,
                prompt=prompt,
                verifier=verifier,
                solution=solution,
                verifier_output=verifier_output,
            )
            if result:
                if self.global_config.generate_patches:
                    result.repaired_source = source_file.get_diff(
                        self.original_source_file
                    )

                return result

        return FixCodeCommandResult(False, self._config.max_attempts, None)

    def _attempt_repair(
        self,
        attempt: int,
        solution_generator: SolutionGenerator,
        prompt: PromptTemplate,
        solution: Solution,
        verifier: ESBMC,
        verifier_output: ESBMCOutput,
    ) -> tuple[FixCodeCommandResult | None, ESBMCOutput]:
        source_file: SourceFile = solution.files[0]

        # Generate AI solution
        with self.anim("Generating Solution... Please Wait"):
            llm_solution = solution_generator.generate_solution(
                initial_message_prompt=prompt,
                solution=solution,
                verifier_output=verifier_output,
            )

            # Update the source file state
            source_file.content = llm_solution

        solution = solution.save_temp()

        # Pass to ESBMC, a workaround is used where the file is saved
        # to a temporary location since ESBMC needs it in file format.
        with self.anim("Verifying with ESBMC... Please Wait"):
            verifier_output = verifier.verify_source(solution=solution)
            assert isinstance(verifier_output, ESBMCOutput)

        # Solution found
        if verifier_output.return_code == 0:

            self.logger.info("Successfully verified code")

            # Check if an output directory is specified and save to it
            if self.global_config.solution.output_dir:
                output_path: Path = (
                    self.global_config.solution.output_dir / source_file.file_path.name
                )
                if self.global_config.generate_patches:
                    output_path = output_path.parent / (output_path.name + ".patch")
                    source_file.save_diff(output_path, self.original_source_file)
                else:
                    source_file.save_file(output_path)
            return FixCodeCommandResult(True, attempt, source_file.content), verifier_output

        # Failure case
        if attempt != self._config.max_attempts:
            self.logger.info(
                f"Failure {attempt}/{self._config.max_attempts}: Retrying..."
            )
        else:
            self.logger.info(
                f"Failure {attempt}/{self._config.max_attempts}: Exiting..."
            )

        return None, verifier_output



# Author: Yiannis Charalambous


from typing import Literal, override
from pathlib import Path

from langchain_core.load.serializable import Serializable
from pydantic import Field

from esbmc_ai.program_trace import ProgramTrace, CounterexampleProgramTrace


class Issue(Serializable):
    """Generic issue/error representation.

    Uses stack_trace as the single source of truth for location information.
    All issues must have at least one trace point describing the error location.
    Simple errors have a single trace point, while complex errors have multiple
    trace points.
    """

    error_type: str = Field(description="Category/type of error.")
    """Category/type of error."""
    message: str = Field(description="Error description.")
    """Error description."""
    stack_trace: list[ProgramTrace] = Field(
        min_length=1,
        description="Stack trace as structured data. At least one trace point required.",
    )
    """Stack trace as structured data. At least one trace point required."""
    severity: Literal["error", "warning", "info"] = Field(
        default="error", description="Severity level."
    )
    """Severity level."""

    @classmethod
    @override
    def is_lc_serializable(cls) -> bool:
        """Enable langchain serialization."""
        return True

    # Convenience properties
    # Note: All properties derive from the last trace point (stack_trace[-1]) as this
    # represents the point of failure in the stack trace. Earlier trace points show
    # the call chain leading to the error.

    @property
    def severity_level(self) -> int:
        """Returns the severity as an int."""
        match self.severity:
            case "info":
                return 0
            case "warning":
                return 1
            case "error":
                return 2

    @property
    def file_path(self) -> Path:
        """Path to file with issue (derived from last trace point)."""
        return self.stack_trace[-1].path

    @property
    def line_index(self) -> int:
        """Line index where the error occurred (derived from last trace point, 0-based)"""
        return self.stack_trace[-1].line_idx

    @property
    def line_number(self) -> int:
        """Line number where error occurred (derived from last trace point, 1-based)."""
        return self.stack_trace[-1].line_idx + 1

    @property
    def function_name(self) -> str | None:
        """Function name where error occurred (derived from last trace point)."""
        return self.stack_trace[-1].name

    @property
    def stack_trace_formatted(self) -> str:
        """Returns a formatted string representation of the stack trace.

        Format: Each trace on a new line showing function name and location.
        Example:
            at main in file.c:15
            at helper in file.c:42
        """
        lines = []
        for trace in self.stack_trace:
            func_name = trace.name if trace.name else "<unknown>"
            line_num = trace.line_idx + 1  # Convert to 1-based
            lines.append(f"\tat {func_name} in {trace.path}:{line_num}")
        return "\n".join(lines)


class VerifierIssue(Issue):
    """Verifier-specific issue with additional verification data.

    This class extends Issue to support verifiers like ESBMC that can provide
    counterexamples in addition to stack traces:

    - stack_trace: Traditional function call stack showing the path to the error
    - counterexample: Program state trace showing variable values and execution
      states that lead to the bug. This is specific to model checkers and formal
      verification tools like ESBMC.

    Note: Not all verifiers support counterexamples (e.g., pytest only provides
    stack traces). Use this class only when counterexample data is available.
    """

    counterexample: list[CounterexampleProgramTrace] = Field(
        description="Counterexample demonstrating bug."
    )
    """Counterexample demonstrating bug."""

    @property
    def counterexample_formatted(self) -> str:
        """Returns a formatted string representation of the counterexample trace.

        Format: Each trace on a new line showing function name, location, and
        assignment (if available).
        Example:
            State 0: at main in file.c:15
              dist = { 0, 0, 0, 0, 0 }
            State 1: at helper in file.c:42
              dist[0] = 2147483647 (01111111 11111111 11111111 11111111)
        """
        lines = []
        for trace in self.counterexample:
            func_name = trace.name if trace.name else "<unknown>"
            line_num = trace.line_idx + 1  # Convert to 1-based
            lines.append(
                f"\tState {trace.trace_index}: at {func_name} in {trace.path}:{line_num}"
            )
            # Add assignment information if available
            if trace.assignment:
                lines.append(f"\t\t{trace.assignment}")
        return "\n".join(lines)



# Author: Yiannis Charalambous

"""Module contains class for keeping track and managing built-in base
components."""

from types import MappingProxyType
from typing import cast

import structlog

from esbmc_ai.base_component import BaseComponent, BaseComponentConfig
from esbmc_ai.chat_command import ChatCommand
from esbmc_ai.singleton import SingletonMeta
from esbmc_ai.verifiers.base_source_verifier import BaseSourceVerifier
from esbmc_ai.log_utils import LogCategories


class ComponentManager(metaclass=SingletonMeta):
    """Class for keeping track of and initializing local components.

    Local components are classes derived from BaseComponent that use base
    component features (maybe for readability). Built-in commands, built-in
    verifiers for example.

    Manages all the verifiers that are used. Can get the appropriate one based
    on the config."""

    def __init__(self) -> None:
        self._logger: structlog.stdlib.BoundLogger = structlog.get_logger(
            self.__class__.__name__
        ).bind(category=LogCategories.SYSTEM)

        # Internal storage
        self._builtin_commands: dict[str, ChatCommand] = {}
        self._addon_commands: dict[str, ChatCommand] = {}
        self._builtin_verifiers: dict[str, BaseSourceVerifier] = {}
        self._addon_verifiers: dict[str, BaseSourceVerifier] = {}
        self._verifier: BaseSourceVerifier | None = None

        # Cached combined dictionaries for efficiency
        self._all_commands_cache: dict[str, ChatCommand] | None = None
        self._all_verifiers_cache: dict[str, BaseSourceVerifier] | None = None
        self._all_components_cache: dict[str, BaseComponent] | None = None
        self._builtin_components_cache: dict[str, BaseComponent] | None = None
        self._addon_components_cache: dict[str, BaseComponent] | None = None

    def _invalidate_caches(self) -> None:
        """Invalidate all cached combined dictionaries."""
        self._all_commands_cache = None
        self._all_verifiers_cache = None
        self._all_components_cache = None
        self._builtin_components_cache = None
        self._addon_components_cache = None

    # ====================
    # Commands API
    # ====================

    @property
    def builtin_commands(self) -> MappingProxyType[str, ChatCommand]:
        """Returns a read-only view of all builtin commands."""
        return MappingProxyType(self._builtin_commands)

    @property
    def addon_commands(self) -> MappingProxyType[str, ChatCommand]:
        """Returns a read-only view of all addon commands."""
        return MappingProxyType(self._addon_commands)

    @property
    def commands(self) -> MappingProxyType[str, ChatCommand]:
        """Returns a read-only view of all commands (builtin + addon)."""
        if self._all_commands_cache is None:
            self._all_commands_cache = self._builtin_commands | self._addon_commands
        return MappingProxyType(self._all_commands_cache)

    @property
    def command_names(self) -> list[str]:
        """Returns a list of all command names."""
        return list(self.commands.keys())

    def add_command(self, command: ChatCommand, builtin: bool = False) -> None:
        """Add a command to the manager."""
        if builtin:
            self._builtin_commands[command.name] = command
        else:
            self._addon_commands[command.name] = command
        self._invalidate_caches()

    def remove_command(self, name: str) -> bool:
        """Remove a command by name. Returns True if removed, False if not found."""
        removed = False
        if name in self._builtin_commands:
            del self._builtin_commands[name]
            removed = True
        if name in self._addon_commands:
            del self._addon_commands[name]
            removed = True
        if removed:
            self._invalidate_caches()
        return removed

    def get_command(self, name: str) -> ChatCommand | None:
        """Get a command by name."""
        return self.commands.get(name)

    def set_builtin_commands(self, builtin_commands: list[ChatCommand]) -> None:
        """Sets the builtin commands, replacing any existing builtin commands."""
        self._builtin_commands = {cmd.command_name: cmd for cmd in builtin_commands}
        self._invalidate_caches()

    # ====================
    # Verifiers API
    # ====================

    @property
    def builtin_verifiers(self) -> MappingProxyType[str, BaseSourceVerifier]:
        """Returns a read-only view of all builtin verifiers."""
        return MappingProxyType(self._builtin_verifiers)

    @property
    def addon_verifiers(self) -> MappingProxyType[str, BaseSourceVerifier]:
        """Returns a read-only view of all addon verifiers."""
        return MappingProxyType(self._addon_verifiers)

    @property
    def verifiers(self) -> MappingProxyType[str, BaseSourceVerifier]:
        """Returns a read-only view of all verifiers (builtin + addon)."""
        if self._all_verifiers_cache is None:
            self._all_verifiers_cache = self._builtin_verifiers | self._addon_verifiers
        return MappingProxyType(self._all_verifiers_cache)

    @property
    def verifier(self) -> BaseSourceVerifier:
        """Returns the currently selected verifier."""
        assert self._verifier, "Verifier is not set..."
        return self._verifier

    @verifier.setter
    def verifier(self, value: BaseSourceVerifier) -> None:
        assert value.name in self.verifiers, f"Unregistered verifier set: {value.name}"
        self._verifier = value

    def add_verifier(self, verifier: BaseSourceVerifier, builtin: bool = True) -> None:
        """Adds a verifier."""
        from esbmc_ai.config import Config

        if builtin:
            self._builtin_verifiers[verifier.name] = verifier
        else:
            self._addon_verifiers[verifier.name] = verifier
        verifier.global_config = Config()
        self._invalidate_caches()

    def remove_verifier(self, name: str) -> bool:
        """Remove a verifier by name. Returns True if removed, False if not found."""
        removed = False
        if name in self._builtin_verifiers:
            del self._builtin_verifiers[name]
            removed = True
        if name in self._addon_verifiers:
            del self._addon_verifiers[name]
            removed = True
        if removed:
            self._invalidate_caches()
        return removed

    def set_verifier_by_name(self, value: str) -> None:
        """Set the active verifier by name."""
        self.verifier = self.verifiers[value]
        self._logger.info(f"Main Verifier: {value}")

    def get_verifier(self, value: str) -> BaseSourceVerifier | None:
        """Get a verifier by name."""
        return self.verifiers.get(value)

    # ====================
    # Components API
    # ====================

    @property
    def builtin_components(self) -> MappingProxyType[str, BaseComponent]:
        """Returns a read-only view of all builtin components (commands + verifiers)."""
        if self._builtin_components_cache is None:
            self._builtin_components_cache = cast(
                dict[str, BaseComponent], self._builtin_commands
            ) | cast(dict[str, BaseComponent], self._builtin_verifiers)
        return MappingProxyType(self._builtin_components_cache)

    @property
    def addon_components(self) -> MappingProxyType[str, BaseComponent]:
        """Returns a read-only view of all addon components (commands + verifiers)."""
        if self._addon_components_cache is None:
            self._addon_components_cache = cast(
                dict[str, BaseComponent], self._addon_commands
            ) | cast(dict[str, BaseComponent], self._addon_verifiers)

        assert self._addon_components_cache is not None
        return MappingProxyType(self._addon_components_cache)

    @property
    def components(self) -> MappingProxyType[str, BaseComponent]:
        """Returns a read-only view of all components (builtin + addon commands + verifiers)."""
        if self._all_components_cache is None:
            self._all_components_cache = (
                cast(dict[str, BaseComponent], self._builtin_commands)
                | cast(dict[str, BaseComponent], self._addon_commands)
                | cast(dict[str, BaseComponent], self._builtin_verifiers)
                | cast(dict[str, BaseComponent], self._addon_verifiers)
            )
        return MappingProxyType(self._all_components_cache)

    def get_component(self, name: str) -> BaseComponent | None:
        """Get a component by name (command or verifier)."""
        return self.components.get(name)

    def load_component_config(self, component: BaseComponent, builtin: bool) -> None:
        """Load component-specific configuration.

        Component configs are loaded automatically via BaseComponentConfig.settings_customise_sources(),
        which loads from TOML, env vars, and .env files.
        """
        try:
            # Check if component has a config instance set
            if component.config is None:
                raise NotImplementedError()

            # Get the config class from the existing instance
            config_class: type[BaseComponentConfig] = type(component.config)

            # Instantiate the config - settings_customise_sources will handle
            # actual loading. Pass component_name via _component_name parameter
            # and builtin via _builtin so BaseComponentConfig can use it for
            # TOML table header. These are not defined fields but captured via
            # extra="ignore" and used in settings_customise_sources.
            loaded_config = config_class(  # type: ignore[call-arg]
                _component_name=component.name, _builtin=builtin
            )

            # Replace the component's config with the loaded one
            component.config = loaded_config
            self._logger.debug(f"Loaded component config for {component.name}")

        except NotImplementedError:
            self._logger.debug(f"No config for component: {component.name}")

        except Exception as e:
            self._logger.error(
                f"Failed to load config for component {component.name}: {e}"
            )



# Author: Yiannis Charalambous

from typing import Any
from typing_extensions import override
from pydantic.fields import FieldInfo
from pydantic import BaseModel

from esbmc_ai.config import Config
from esbmc_ai.chat_command import ChatCommand
from esbmc_ai.command_result import CommandResult
from esbmc_ai.component_manager import ComponentManager
from esbmc_ai.base_component import BaseComponent


class HelpConfigCommand(ChatCommand):
    """Command that prints the help messages of the acceptable config fields."""

    def __init__(self) -> None:
        super().__init__(
            command_name="help-config",
            help_message="Print information about the config fields.",
        )

    @staticmethod
    def _print_config_field(
        field_name: str, field_info: FieldInfo, level: int = 0
    ) -> None:
        """Print information about a single config field."""
        # Create indentation based on level
        indent = "\t" * (level + 1)

        # Get default value and format it
        default_value = field_info.default
        if isinstance(default_value, str) and len(default_value) > 30:
            default_value = default_value[:30] + "..."

        print(f"{indent}* {field_name}:")

        if field_info.description:
            print(f"{indent}  Description: {field_info.description}")

        # Check if this field's annotation is a BaseModel type
        field_annotation = field_info.annotation
        if field_annotation and hasattr(field_annotation, "__origin__"):
            # Handle generic types like list, dict, etc.
            field_annotation = field_annotation.__origin__

        # Check if the field type is a BaseModel subclass
        if (
            field_annotation
            and isinstance(field_annotation, type)
            and issubclass(field_annotation, BaseModel)
        ):

            print(f"{indent}  Nested fields:")
            # Recursively print fields of the BaseModel
            for (
                nested_field_name,
                nested_field_info,
            ) in field_annotation.model_fields.items():
                HelpConfigCommand._print_config_field(
                    nested_field_name, nested_field_info, level + 1
                )
        else:
            # Print default value for non-BaseModel fields
            if field_info.default is not None:
                print(f"{indent}  Default: {default_value}")

            if hasattr(field_info, "alias") and field_info.alias:
                print(f"{indent}  Alias: {field_info.alias}")

        print()

    @staticmethod
    def _print_component_config_fields(
        component_name: str, component: BaseComponent, level: int = 0
    ) -> bool:
        """Print config fields for a component if it has any.

        Returns True if the component has config fields, False otherwise.
        """
        try:
            # Try to access the component's config
            config = component.config
            if config is None:
                return False

            # Get the config class and its fields
            config_class = type(config)
            config_fields = config_class.model_fields

            if not config_fields:
                return False

            indent = "\t" * level
            print(f"{indent}{component_name}:")

            for field_name, field_info in config_fields.items():
                if not getattr(field_info, "exclude", False):
                    HelpConfigCommand._print_config_field(field_name, field_info, level)

            return True

        except (NotImplementedError, RuntimeError, AttributeError):
            # Component doesn't have config
            return False

    @staticmethod
    def _print_components_section(
        title: str, components: dict[str, BaseComponent]
    ) -> None:
        """Print a section for a group of components (builtin or addon)."""
        # First, collect components that have config
        components_with_config: list[tuple[str, BaseComponent]] = []

        for name, component in components.items():
            try:
                # Try to access config to check if it exists
                if component.config is not None:
                    components_with_config.append((name, component))
            except (NotImplementedError, RuntimeError, AttributeError):
                # Component doesn't have config, skip it
                continue

        if not components_with_config:
            # No components with config in this section
            return

        print(title)
        print()

        for name, component in components_with_config:
            HelpConfigCommand._print_component_config_fields(name, component)

    @override
    def execute(self) -> CommandResult | None:
        print("ESBMC-AI Config Fields:")
        print()

        # Get all fields from the Config model
        config_fields = Config.model_fields

        for field_name, field_info in config_fields.items():
            # Skip excluded fields (like command_name and config_file)
            if not getattr(field_info, "exclude", False):
                self._print_config_field(field_name, field_info)

        from esbmc_ai.log_utils import print_horizontal_line

        print_horizontal_line()

        # Get component manager to access builtin and addon components
        component_manager = ComponentManager()

        # Print builtin component config fields
        self._print_components_section(
            "Builtin Component Config Fields:",
            dict(component_manager.builtin_components),
        )

        print_horizontal_line()

        # Print addon component config fields
        self._print_components_section(
            "Addon Component Config Fields:", dict(component_manager.addon_components)
        )

        return None



# Author: Yiannis Charalambous

from pathlib import Path
from pydantic import BaseModel, Field


class ProgramTrace(BaseModel):
    """Contains information about traces in source code."""

    trace_index: int = Field()
    """The index position of this trace point in the trace stack."""
    path: Path = Field()
    """The source file of this trace. May not exist for compilation errors. May
    not be relative as it may be a system file."""
    name: str | None = Field(default=None)
    """The name of the symbol pointed to by the trace, if applicable."""
    line_idx: int = Field()
    """The location of the trace (0-based)."""


class CounterexampleProgramTrace(ProgramTrace):
    """Program trace with assignment information from counterexample states.

    This class extends ProgramTrace to capture the variable assignments that
    led to a verification failure. Used by verifiers like ESBMC that provide
    counterexample traces showing the state of variables at each step.
    """

    assignment: str | None = Field(default=None)
    """The assignment statement(s) for this trace state. Contains the variable
    assignments from the counterexample, e.g., 'dist = { 0, 0, 0, 0, 0 }' or
    'dist[0] = 2147483647 (01111111 11111111 11111111 11111111)'. May be None
    if the state has no assignment information."""



# Author: Yiannis Charalambous

"""This module contains different chat interfaces. Along with `BaseChatInterface`
that provides necessary boilet-plate for implementing an LLM based chat."""

from .template_key_provider import (
    TemplateKeyProvider,
    GenericTemplateKeyProvider,
    OracleTemplateKeyProvider,
)

from .key_template_renderer import KeyTemplateRenderer

__all__ = [
    "TemplateKeyProvider",
    "GenericTemplateKeyProvider",
    "OracleTemplateKeyProvider",
    "KeyTemplateRenderer",
]



"""This module contains built-in commands that can be executed by ESBMC-AI.
NOTE: Do not add any that are not builtin and will be exposed to the main program."""

from .exit_command import ExitCommand
from .help_command import HelpCommand
from .help_config import HelpConfigCommand
from .fix_code_command import FixCodeCommand
from .debug_config import DebugConfigViewCommand
from .license_command import LicenseCommand

__all__ = [
    "ExitCommand",
    "HelpCommand",
    "HelpConfigCommand",
    "FixCodeCommand",
    "DebugConfigViewCommand",
    "LicenseCommand",
]



# Author: Yiannis Charalambous

"""This module contains code regarding configuring and loading addon modules."""

import traceback
import sys
import importlib
import structlog

from esbmc_ai.base_component import BaseComponent
from esbmc_ai.chat_command import ChatCommand
from esbmc_ai.verifiers.base_source_verifier import BaseSourceVerifier
from esbmc_ai.config import Config
from esbmc_ai.singleton import SingletonMeta


class AddonLoader(metaclass=SingletonMeta):
    """The addon loader manages loading addon modules and initializing them."""

    addon_prefix: str = "addons"

    def __init__(self, config: Config | None = None) -> None:
        self._logger: structlog.stdlib.BoundLogger = structlog.get_logger(
            self.__class__.__name__
        )
        self._logger.debug("Loading Addons...")

        assert config
        self._config: Config = config

        # Keeps track of the addons that have been loaded.
        self._loaded_addons: dict[str, BaseComponent] = {}

        # Dev mode: Ensure the current directory is in sys.path in order for
        # relative addon modules to be imported (used for dev purposes).
        if self._config.dev_mode and "" not in sys.path:
            sys.path.insert(0, "")

        # Load the config fields.
        if self._config.addon_modules:
            print("Loading Addons:")

        for m in self._config.addon_modules:
            addons: list[BaseComponent] = self.load_addons_module(m)
            for addon in addons:
                print(f"\t* {addon.name} by {addon.authors}")

    @property
    def chat_command_addons(self) -> dict[str, ChatCommand]:
        """Returns all the addon chat commands."""
        return {
            addon_name: addon
            for addon_name, addon in self.loaded_addons.items()
            if isinstance(addon, ChatCommand)
        }

    @property
    def chat_command_addon_names(self) -> list[str]:
        """Returns all the addon chat command names."""
        return list(
            addon.name
            for addon in self.loaded_addons.values()
            if isinstance(addon, ChatCommand)
        )

    @property
    def verifier_addons(self) -> dict[str, BaseSourceVerifier]:
        """Returns all the addon verifiers."""
        return {
            addon_name: addon
            for addon_name, addon in self.loaded_addons.items()
            if isinstance(addon, BaseSourceVerifier)
        }

    @property
    def verifier_addon_names(self) -> list[str]:
        """Returns all the addon verifier names."""
        return list(
            addon.name
            for addon in self.loaded_addons
            if isinstance(addon, BaseSourceVerifier)
        )

    @property
    def loaded_addons(self) -> dict[str, BaseComponent]:
        return self._loaded_addons

    def get_addon_by_name(self, name: str) -> BaseComponent | None:
        """Returns an addon by the addon name defined in __init__."""
        for addon in self.loaded_addons.values():
            if addon.name == name:
                return addon
        return None

    def load_addons_module(self, module_name: str) -> list[BaseComponent]:
        """Loads an addon, needs to expose a BaseComponent.

        Will import addon modules that exist and iterate through the exposed
        attributes, will then get all available exposed classes and store them."""

        result: list = []
        try:
            m = importlib.import_module(module_name)
            for exposed_class_name in getattr(m, "__all__"):
                # Get the class from the __all__
                exposed_class = getattr(m, exposed_class_name)
                # Check if valid addon type and import
                if issubclass(exposed_class, BaseComponent):
                    addon: BaseComponent = self.init_base_component(exposed_class)
                    self._loaded_addons[addon.name] = addon
                    result.append(addon)

        except ModuleNotFoundError as e:
            self._logger.error("error while loading module: traceback:")
            traceback.print_tb(e.__traceback__)
            self._logger.error(f"could not import module: {module_name}: {e}")
            sys.exit(1)
        except AttributeError as e:
            self._logger.error(f"module {module_name} is invalid: {e}")
            sys.exit(1)
        return result

    def init_base_component(self, t: type[BaseComponent]) -> BaseComponent:
        # Initialize class.
        addon: BaseComponent = t.create()
        self._logger.debug(f"Loading addon: {addon.__class__.__name__}")

        # Register config with modules
        addon.global_config = self._config

        # Load component-specific configuration using ComponentManager
        from esbmc_ai.component_manager import ComponentManager

        ComponentManager().load_component_config(addon, builtin=False)

        return addon



# Author: Yiannis Charalambous

"""Contains the license command."""

from typing import Any
from typing_extensions import override

from esbmc_ai.chat_command import ChatCommand


class LicenseCommand(ChatCommand):
    """Command that displays license and contribution information for ESBMC-AI."""

    def __init__(self) -> None:
        super().__init__(
            command_name="license",
            help_message="Display license and contribution information.",
        )

    @override
    def execute(self) -> Any:
        print(
            """
╔════════════════════════════════════════════════════════════════╗
║                  ESBMC-AI License Information                  ║
╚════════════════════════════════════════════════════════════════╝

ESBMC-AI is offered under a DUAL-LICENSE model:

📖 Open Source License - AGPL-3.0
    You are free to use, modify, and distribute this software under the GNU 
    Affero General Public License v3.0, provided you:
    • Disclose the source code of your application
    • Distribute under the same AGPL-3.0 license
    • Make source available to all network users

💼 Commercial License

    For projects where the terms of the AGPL-3.0 are not suitable, such as in 
    proprietary or closed-source applications, we offer a separate commercial 
    license.

    Contact: contact@uominnovationfactory.com

📝 Contributing
    All contributions require signing our Contributor License Agreement (CLA) 
    based on the Apache Software Foundation CLA.

For full details:
    • License: https://github.com/esbmc/esbmc-ai/blob/master/LICENSE
    • CLA: https://github.com/esbmc/esbmc-ai/blob/master/CLA.md
    • Wiki: https://esbmc.github.io/esbmc-ai/contributing
"""
        )



# Author: Yiannis Charalambous

from abc import abstractmethod
from functools import cached_property
from pathlib import Path
from typing import override

from langchain_core.load.serializable import Serializable
from pydantic import Field

from esbmc_ai.program_trace import ProgramTrace
from esbmc_ai.issue import Issue


class VerifierOutput(Serializable):
    """Class that represents the verifier output. All properties can be accessed
    directly in templates.

    This class provides both structured issue information (via the issues list)
    and convenient property access for single-error scenarios.
    """

    return_code: int
    """The return code of the verifier."""
    output: str
    """The output of the verifier."""
    issues: list[Issue] = Field(default_factory=list)
    """List of issues/errors found during verification."""
    duration: float | None = None
    """Execution time in seconds."""

    @classmethod
    @override
    def is_lc_serializable(cls) -> bool:
        """Enable langchain serialization."""
        return True

    @property
    def issue_count(self) -> int:
        """Total number of issues found."""
        return len(self.issues)

    @property
    @abstractmethod
    def successful(self) -> bool:
        """If the verification was successful."""
        raise NotImplementedError()

    # Convenience methods

    @cached_property
    def primary_issue(self) -> Issue:
        """Returns the primary issue to address - the issue with highest severity.

        When multiple issues have the same severity level, returns the first one
        (which typically represents the root cause in execution order).
        """
        assert self.issues, "No issues found in verifier output"
        return max(self.issues, key=lambda issue: issue.severity_level)

    @property
    def error_line(self) -> int:
        """Returns the line number where the error occurred."""
        return self.issues[0].line_number

    @property
    def error_line_idx(self) -> int:
        """Returns the line index where the error occurred (0-based)."""
        return self.issues[0].line_index

    @property
    def error_message(self) -> str | None:
        """first error message from issues list."""
        return self.issues[0].message if self.issues else None

    @property
    def error_file(self) -> Path | None:
        """first error file path from issues list."""
        return self.issues[0].file_path if self.issues else None

    @property
    def error_type(self) -> str:
        """Returns a string of the type of error found by the verifier output."""
        return self.issues[0].error_type

    @property
    def stack_trace(self) -> list[ProgramTrace]:
        """Gets the stack trace that points to the error."""
        return self.issues[0].stack_trace



# Author: Yiannis Charalambous

from pathlib import Path

from esbmc_ai.issue import Issue
from esbmc_ai.program_trace import ProgramTrace


class ClangOutputParser:
    """Parser for Clang/GCC diagnostic output format.

    Handles compilation errors and warnings in the standard format:
    <filename>:<line>:<col>: <error|warning|note>: <message>
    <source line>
    <indicator line (e.g., ^ pointing to error)>
    """

    @staticmethod
    def _parse_diagnostic_lines(output: str) -> list[tuple[str, int, str]]:
        """Extract diagnostic entries from compiler output.

        Args:
            output: Compiler output text containing diagnostics

        Returns:
            List of (filename, line_number, message) tuples for each error
        """
        output_split: list[str] = output.splitlines(True)

        # Find the first line not starting with "Parsing" (ESBMC-specific prefix)
        # Skip first 2 lines as they typically contain program header
        start_line: int = 2
        for idx, line in enumerate(output_split[start_line:], start=start_line):
            if not line.startswith("Parsing"):
                start_line = idx
                break

        diagnostics: list[tuple[str, int, str]] = []
        idx: int = start_line
        while idx < len(output_split):
            line = output_split[idx]
            line_split: list[str] = line.split(":")

            # Expected format: filename:line:col:type:message
            if len(line_split) == 5 and line_split[3].strip() == "error":
                filename: str = line_split[0]
                line_number: int = int(line_split[1])
                # col: int = int(line_split[2])  # Column not currently used
                message: str = line_split[4].strip()
                diagnostics.append((filename, line_number, message))

            # Skip 3 lines per diagnostic entry:
            # Line 1: "file:line:col: error: message"
            # Line 2: The actual source code line
            # Line 3: Error indicator (e.g., "^" or "~~~~~" pointing to error)
            idx += 3

        return diagnostics

    @staticmethod
    def parse_diagnostics(output: str) -> list[Issue]:
        """Parse Clang/GCC diagnostic format into Issue objects.

        Args:
            output: Compiler output text containing diagnostics

        Returns:
            List of Issue objects for each error found
        """
        issues: list[Issue] = []

        for filename, line_number, message in ClangOutputParser._parse_diagnostic_lines(
            output
        ):
            # Create a single-point trace for the compilation error
            trace: ProgramTrace = ProgramTrace(
                trace_index=0,
                path=Path(filename),
                line_idx=line_number - 1,  # Convert to 0-based indexing
            )

            issues.append(
                Issue(
                    error_type="Compilation Error",
                    message=message,
                    stack_trace=[trace],
                    severity="error",
                )
            )

        return issues



# Author: Yiannis Charalambous

from enum import Enum


class LogCategories(Enum):
    NONE = "NONE"
    ALL = "ALL"
    SYSTEM = "ESBMC_AI"
    VERIFIER = "VERIFIER"
    COMMAND = "COMMAND"
    CONFIG = "CONFIG"
    CHAT = "CHAT"



# Author: Yiannis Charalambous

"""Contains mappings for template values and filters that are used by
KeyTemplateRenderer prompt templates."""

from typing import Any

from esbmc_ai.issue import VerifierIssue


def get_func_mapping():
    return _func_mapping


def _is_verifier_issue(obj: Any) -> bool:
    return isinstance(obj, VerifierIssue)


_func_mapping: dict[str, Any] = {
    "is_verifier_issue": _is_verifier_issue,
}


# TODO Figure out how to integrate into KeyTemplateRenderer

# def get_filter_mapping():
#     return _filter_mapping


# def _isinstance_filter(obj: object, classinfo: type) -> bool:
#     return isinstance(obj, classinfo)


# _filter_mapping: dict[str, Callable] = {
#     "isinstance": _isinstance_filter,
# }



# Author: Yiannis Charalambous

from typing import Any, override
from esbmc_ai.chat_command import ChatCommand
from esbmc_ai.command_result import CommandResult
from esbmc_ai.component_manager import ComponentManager


class DebugConfigViewCommand(ChatCommand):
    """Displays all loaded config values."""

    def __init__(self) -> None:
        super().__init__(
            command_name="debug-view-config",
            authors="",
            help_message="Used for debug to view the current config state.",
        )

    @staticmethod
    def _format_value(value: Any, indent: str = "  ") -> None:
        """Format and print a config value with appropriate formatting."""
        if isinstance(value, dict):
            if value:
                for k, v in value.items():
                    print(f"{indent}{k}: {v}")
            else:
                print(f"{indent}(empty)")
        elif isinstance(value, (list, tuple)):
            if value:
                for item in value:
                    print(f"{indent}- {item}")
            else:
                print(f"{indent}(empty)")
        else:
            print(f"{indent}{value}")

    def _print_config_section(
        self, config_dict: dict[str, Any], indent: str = ""
    ) -> None:
        """Print a config dictionary with formatted field names."""
        for field_name, value in config_dict.items():
            display_name = field_name.replace("_", " ").title()
            print(f"\n{indent}{display_name}:")
            self._format_value(value, indent + "  ")

    @override
    def execute(self) -> CommandResult | None:
        print("\n" + "=" * 80)
        print("GLOBAL CONFIGURATION")
        print("=" * 80)

        self._print_config_section({"Config File": self.global_config.config_file})
        self._print_config_section(self.global_config.model_dump())

        print("\n" + "=" * 80)
        print("BUILTIN COMPONENT CONFIGURATIONS")
        print("=" * 80)

        cm = ComponentManager()
        builtin_components = cm.builtin_components
        if not builtin_components:
            print("\n(No builtin components)")
        else:
            for name, component in builtin_components.items():
                print(f"\n{name}:")
                try:
                    if component.config:
                        self._print_config_section(
                            component.config.model_dump(), indent="  "
                        )
                    else:
                        print("  (no config)")
                except NotImplementedError:
                    print("  (no config)")

        print("\n" + "=" * 80)
        print("ADDON COMPONENT CONFIGURATIONS")
        print("=" * 80)

        addon_components = cm.addon_components
        if not addon_components:
            print("\n(No addon components)")
        else:
            for name, component in addon_components.items():
                print(f"\n{name}:")
                try:
                    if component.config:
                        self._print_config_section(
                            component.config.model_dump(), indent="  "
                        )
                    else:
                        print("  (no config)")
                except NotImplementedError:
                    print("  (no config)")

        print("\n" + "=" * 80 + "\n")

        return None



# Author: Yiannis Charalambous

"""Keeps track of all the source files that ESBMC-AI is targeting."""

from os import walk
from os.path import commonpath
from pathlib import Path
from subprocess import PIPE, STDOUT, run, CompletedProcess
from tempfile import NamedTemporaryFile, TemporaryDirectory
from shutil import copytree
from typing import Any, Literal, override

from langchain_core.language_models import BaseChatModel
from langchain_core.load.serializable import Serializable
from pydantic import Field, PrivateAttr
import lizard
from structlog.stdlib import get_logger

from esbmc_ai.log_utils import LogCategories, get_log_level, print_horizontal_line

_SourceFileFormatStyles = Literal["markdown", "xml", "plain"]


class SolutionIntegrityError(Exception):
    """Raised when the solution disk integrity check fails."""

    def __init__(self, files: list["SourceFile"]):
        assert all(
            isinstance(f, SourceFile) for f in files
        ), "Accept only a list of SourceFiles"
        self.invalid_files: list[SourceFile] = [
            f for f in files if not f.verify_file_integrity()
        ]
        super().__init__(self._build_message())

    def _build_message(self):
        files_list = "\n\t* " + "\n\t* ".join(
            str(f.file_path) for f in self.invalid_files
        )
        return (
            "Solution disk integrity check failed. There are unsaved "
            "changes. Save solution to a temporary location on disk.\n"
            "The following files are invalid:" + files_list
        )


class SourceFile(Serializable):
    """Represents a source file in the Solution. Contains methods to manipulate
    and get information about different versions."""

    file_path: Path = Field(description="Absolute file path")
    content: str = Field(description="File content")

    @staticmethod
    def apply_line_patch(source_code: str, patch: str, start: int, end: int) -> str:
        """Applies a patch to the source code.

        To replace a single line, start and end are equal.

        Args:
            * source_code - The source code to apply the patch to.
            * patch - Can be a line or multiple lines but will replace the start and
            end region defined.
            * start - Line index to mark start of replacement.
            * end - Marks the end of the region where the patch will be applied to.
            End is non-inclusive."""
        assert (
            start <= end
        ), f"start ({start}) needs to be less than or equal to end ({end})"
        lines: list[str] = source_code.splitlines()
        lines = lines[:start] + [patch] + lines[end + 1 :]
        return "\n".join(lines)

    @staticmethod
    def load(file_path: Path) -> "SourceFile":
        """Creates a new source file by loading the content from disk."""
        abs_path = file_path.absolute()
        with open(abs_path, "r") as file:
            return SourceFile(file_path=abs_path, content=file.read())

    def __init__(
        self,
        file_path: Path,
        content: str,
        **kwargs: Any,
    ) -> None:
        # Ensure path is absolute
        if not file_path.is_absolute():
            file_path = file_path.absolute()
        super().__init__(
            file_path=file_path,
            content=content,
            **kwargs,
        )

    @override
    def __repr__(self) -> str:
        return f"SourceFile({self.file_path}, valid={self.verify_file_integrity()})"

    @override
    def __str__(self) -> str:
        return f"SourceFile({self.file_path})"

    @property
    def file_extension(self) -> str:
        """Returns the file extension to the file."""
        return self.file_path.suffix

    @property
    def line_count(self) -> int:
        """Returns the total number of lines in the file."""
        return len(self.content.splitlines())

    def get_num_tokens(
        self,
        ai_model: BaseChatModel,
        lower_idx: int | None = None,
        upper_idx: int | None = None,
    ) -> int:
        """Gets the context size this source code."""
        if not lower_idx:
            lower_idx = 0
        if not upper_idx:
            upper_idx = len(self.content)

        assert lower_idx < upper_idx
        return ai_model.get_num_tokens(self.content[lower_idx:upper_idx])

    def get_diff(self, source_file_2: "SourceFile") -> str:
        """Return diff between two SourceFiles."""
        # Save as temp files
        with (
            NamedTemporaryFile(mode="w", delete=False) as file,
            NamedTemporaryFile(mode="w", delete=False) as file_2,
        ):
            file.write(self.content)
            file.flush()
            file_2.write(source_file_2.content)
            file_2.flush()

            cmd = [
                "diff",
                "-u",  # Adds context around the diff text.
                "--label",
                str(source_file_2.file_path),  # Label for first file (original)
                "--label",
                str(self.file_path),  # Label for second file (modified)
                file_2.name,  # Original file first
                file.name,  # Modified file second
            ]
            process: CompletedProcess = run(
                cmd,
                stdout=PIPE,
                stderr=STDOUT,
                check=False,
            )

            # Exit status is 0 if inputs are the same, 1 if different, 2 if trouble.
            # https://askubuntu.com/questions/698784/exit-code-of-diff
            if process.returncode == 2:
                raise ValueError(
                    f"Diff for {file.name} and {file_2.name} failed (exit 2)."
                )

            return process.stdout.decode("utf-8")

    def save_temp_file(self) -> Path:
        """Saves the file in a temporary directory it creates. The temp file is
        placed in a temporary directory"""
        parent_dir: Path = self.file_path.parent
        with TemporaryDirectory(
            prefix="esbmc-ai", suffix="-" + parent_dir.name
        ) as temp_dir:
            return self.save_file(Path(temp_dir) / self.file_path)

    def save_file(self, abs_path: Path) -> Path:
        """Saves the source code file."""

        with open(abs_path, "w") as file:
            file.write(self.content)
            return Path(file.name)

    def save_diff(self, path: Path, source_file: "SourceFile") -> Path:
        """Saves the source file as a patch of the original."""
        diff: str = self.get_diff(source_file)
        with open(path, "w") as file:
            file.write(diff)
        return path

    def verify_file_integrity(self) -> bool:
        """Verifies that this file matches with the file on disk and does not
        contain any differences."""
        if not (self.file_path.exists() and self.file_path.is_file()):
            return False

        with open(self.file_path, "r") as file:
            content: str = str(file.read())

        return content == self.content

    def calculate_cyclomatic_complexity_delta(
        self,
        source_2: "SourceFile",
    ) -> float | None:
        """Calculates the cyclomatic complexity difference between the two source files."""
        try:
            file_1: Path = self.save_temp_file()
            file_2: Path = source_2.save_temp_file()

            cc1 = lizard.analyze_file(file_1)
            cc2 = lizard.analyze_file(file_2)

            return cc1.average_cyclomatic_complexity - cc2.average_cyclomatic_complexity
        except IOError:
            return None

    @property
    def formatted(self) -> str:
        """Default markdown formatting - most common use case for LangChain templates."""
        return self.format_as()

    def format_as(
        self,
        style: _SourceFileFormatStyles = "markdown",
        include_line_numbers: bool = False,
        max_lines: int | None = None,
        working_dir: Path | None = None,
    ) -> str:
        """Flexible formatting with options for use in LangChain prompts.

        Args:
            style: Format style - "markdown", "xml", or "plain"
            include_line_numbers: Add line numbers to the content
            max_lines: Limit content to first N lines (with truncation notice)

        Returns:
            Formatted string representation of the source file
        """
        lang = self.file_extension.strip(".")
        content = self.content

        # Truncate if needed
        if max_lines:
            lines = content.splitlines()[:max_lines]
            content = "\n".join(lines)
            if len(self.content.splitlines()) > max_lines:
                content += (
                    f"\n... ({len(self.content.splitlines()) - max_lines} more lines)"
                )

        # Add line numbers if requested
        if include_line_numbers:
            lines = content.splitlines()
            content = "\n".join(f"{i+1:4d} | {line}" for i, line in enumerate(lines))

        # Format based on style
        show_path: Path = (
            self.file_path.relative_to(working_dir) if working_dir else self.file_path
        )
        if style == "markdown":
            result = f"{show_path}\n```{lang}\n{content}\n```"
        elif style == "xml":
            result = f"<file path='{show_path}'>\n{content}\n</file>"
        elif style == "plain":
            result = f"File: {show_path}\n{content}"

        return result


class Solution(Serializable):
    """Represents a solution, that is a collection of all the source files that
    ESBMC-AI will be involved in analyzing."""

    _files: list[SourceFile] = PrivateAttr(default_factory=list)
    _include_dirs: list[Path] = PrivateAttr(default_factory=list)

    @staticmethod
    def from_paths(
        *paths: Path,
        include_dirs: list[Path] | None = None,
    ) -> "Solution":
        """Creates a solution from one or more paths (files or directories).

        Directories are scanned recursively for all files.

        Args:
            *paths: One or more file or directory paths
            include_dirs: Optional include directories for C/C++ compilation

        Example:
            Solution.from_paths(Path("file1.c"), Path("file2.c"))
            Solution.from_paths(Path("/project"))
            Solution.from_paths(Path("src"), Path("tests/test.c"))
        """
        result: list[Path] = []

        for path in paths:
            abs_path = path.absolute()
            if abs_path.is_file():
                result.append(abs_path)
            elif abs_path.is_dir():
                for dir_path, _, files in walk(abs_path):
                    result.extend(Path(dir_path) / file for file in files)
            else:
                raise ValueError(
                    f"Path does not exist or is neither file nor directory: {path}"
                )

        return Solution(files=result, include_dirs=include_dirs)

    def __init__(
        self,
        files: list[Path] | None = None,
        include_dirs: list[Path] | None = None,
        **kwargs: Any,
    ) -> None:
        """Creates a new solution from file paths."""
        super().__init__(**kwargs)

        # Initialize private attributes
        self._files = []
        self._include_dirs = []

        # If files are loaded
        if files:
            for file_path in files:
                # Convert to absolute path
                abs_path = (
                    file_path if file_path.is_absolute() else file_path.absolute()
                )
                if not abs_path.is_file():
                    raise ValueError(f"Path is not a file: {file_path}")
                self._files.append(SourceFile.load(abs_path))

        if include_dirs:
            for d in include_dirs:
                abs_dir = d if d.is_absolute() else d.absolute()
                if not abs_dir.is_dir():
                    raise ValueError(f"Path is not a directory: {d}")
                self._include_dirs.append(abs_dir)

    @property
    def include_dirs(self) -> list[Path]:
        """Include directories for C based verifiers."""
        return list(self._include_dirs)

    @property
    def files(self) -> list[SourceFile]:
        """Will return a list of the files. Returns by value."""
        return list(self._files)

    @property
    def working_dir(self) -> Path:
        """Computes the common parent directory of all files."""
        if not self._files:
            return Path.cwd()

        paths = [str(f.file_path) for f in self._files]
        common = commonpath(paths)

        # If only one file, commonpath returns the file itself, so use parent
        return Path(common).parent if len(paths) == 1 else Path(common)

    def get_files_by_ext(self, included_ext: list[str]) -> list[SourceFile]:
        """Gets the files that are only specified in the included extensions. File
        extensions that have a . prefix are trimmed so they still work."""
        return [s for s in self.files if s.file_extension.strip(".") in included_ext]

    @property
    def files_list_formatted(self) -> str:
        """Bullet point list of all files - useful for LangChain templates."""
        return "\n".join(f"- {f.file_path}" for f in self.files)

    @property
    def formatted(self) -> str:
        """Default: all files formatted with markdown - useful for LangChain templates."""
        return self.format_as()

    def format_as(
        self,
        style: _SourceFileFormatStyles = "markdown",
        include_line_numbers: bool = False,
        max_lines_per_file: int | None = None,
        separator: str = "\n\n---\n\n",
    ) -> str:
        """Format all source files with options for use in LangChain prompts.

        Args:
            style: Format style - "markdown", "xml", or "plain"
            include_line_numbers: Add line numbers to the content
            max_lines_per_file: Limit each file's content to first N lines
            separator: String to separate files (default: horizontal rule)

        Returns:
            Formatted string representation of all source files
        """
        formatted_files = [
            f.format_as(
                style=style,
                include_line_numbers=include_line_numbers,
                max_lines=max_lines_per_file,
            )
            for f in self.files
        ]
        return separator.join(formatted_files)

    def save_temp(self) -> "Solution":
        """Saves the solution in a temporary directory it creates while preserving
        the file structure."""
        with TemporaryDirectory(
            prefix="esbmc-ai-", suffix="-" + self.working_dir.name, delete=False
        ) as temp_dir:
            return self.save_solution(Path(temp_dir))

    def save_solution(self, path: Path) -> "Solution":
        """Saves the solution to path, preserving relative structure.

        Files are saved relative to their common parent (working_dir)."""
        get_logger().bind(category=LogCategories.SYSTEM).info(
            f"Saving solution to {path}"
        )

        dest_path: Path = path.absolute()
        dest_path.mkdir(parents=True, exist_ok=True)

        # Get common parent to preserve relative structure
        common_parent = self.working_dir

        # Copy individual source files preserving structure
        new_file_paths: list[Path] = []
        for source_file in self.files:
            # Get path relative to common parent
            relative_path = source_file.file_path.relative_to(common_parent)
            new_path: Path = dest_path / relative_path

            # Write new file
            new_path.parent.mkdir(parents=True, exist_ok=True)
            source_file.save_file(new_path)
            new_file_paths.append(new_path)

        # Copy include directories
        new_include_dirs: list[Path] = []
        for d in self.include_dirs:
            new_dir: Path = dest_path / d.name
            copytree(src=d, dst=new_dir, symlinks=True, dirs_exist_ok=True)
            new_include_dirs.append(new_dir)

        return Solution(
            files=new_file_paths,
            include_dirs=new_include_dirs,
        )

    def verify_solution_integrity(self) -> bool:
        """Verifies if the content of the solution match with the files on disk.
        If that's not the case, then the solution should be saved on disk before
        any operations such as using verifiers that require on-disk solution
        access to work."""

        return all(f.verify_file_integrity() for f in self._files)

    def add_source_file(self, source_file: SourceFile) -> None:
        """Adds a source file to the solution."""
        self._files.append(source_file)

    def add_source_files(self, source_files: list[SourceFile]) -> None:
        """Adds multiple source files to the solution"""
        for f in source_files:
            self._files.append(f)

    def load_source_files(self, file_paths: list[Path]) -> None:
        """Loads multiple source files from disk."""
        for f in file_paths:
            assert isinstance(f, Path), f"Invalid type: {type(f)}"
            if f.is_dir():
                for path in f.glob("**/*"):
                    if path.is_file() and path.name:
                        self.load_source_file(path)
            else:
                self.load_source_file(f)

    def load_source_file(self, file_path: Path) -> None:
        """Add a source file to the solution by loading it from disk."""
        abs_path = file_path if file_path.is_absolute() else file_path.absolute()
        self._files.append(SourceFile.load(abs_path))

    def get_diff(self, other: "Solution") -> str:
        """Gets the diff with another solution. The following params are used: -ruN"""

        if not self.verify_solution_integrity():
            self.save_temp()

        if not other.verify_solution_integrity():
            other.save_temp()

        cmd = [
            "diff",
            "-ruN",  # (r)ecursive, (u) additional context, (N) treat absent files as empty
            str(self.working_dir),
            str(other.working_dir),
        ]

        process: CompletedProcess = run(
            cmd,
            stdout=PIPE,
            stderr=STDOUT,
            check=False,
        )

        # Exit status is 0 if inputs are the same, 1 if different, 2 if trouble.
        # https://askubuntu.com/questions/698784/exit-code-of-diff
        if process.returncode == 2:
            raise ValueError(
                f"Diff for {self.working_dir} and {other.working_dir} failed (exit 2)."
            )

        return process.stdout.decode("utf-8")

    def save_diff(self, path: Path, solution: "Solution") -> Path:
        """Saves the solution as a patch of the original."""
        diff: str = self.get_diff(solution)
        with open(path, "w") as file:
            file.write(diff)
        return path

    def resolve(self, file: SourceFile | Path) -> SourceFile | None:
        """Resolve a Path or SourceFile to its corresponding SourceFile in the solution.

        Args:
            file: Either a SourceFile instance or a Path (absolute or relative)

        Returns:
            The SourceFile if found, None otherwise
        """
        if isinstance(file, SourceFile):
            search_path = file.file_path
        else:
            # Convert to absolute path for comparison
            search_path = Path(file).absolute()

        for f in self._files:
            if f.file_path == search_path:
                return f
        return None

    def __contains__(self, file: SourceFile | Path) -> bool:
        """Check if a file is part of this solution. Enables 'in' operator.

        Args:
            file: Either a SourceFile instance or a Path (absolute or relative)

        Returns:
            True if the file is in the solution, False otherwise
        """
        return self.resolve(file) is not None

    def get_file(self, path: Path) -> SourceFile:
        """Get a SourceFile by path, raising KeyError if not found.

        Args:
            path: Either an absolute or relative Path

        Returns:
            The SourceFile

        Raises:
            KeyError: If the path is not in the solution
        """
        result = self.resolve(path)
        if result is None:
            raise KeyError(f"Path not found in solution: {path}")
        return result

    def __add__(self, other: "Solution") -> "Solution":
        """Combine two solutions by merging their files and include directories.

        Args:
            other: Another Solution instance to combine with this one

        Returns:
            A new Solution containing all files and include_dirs from both solutions

        Example:
            solution_c = solution_a + solution_b
        """
        if not isinstance(other, Solution):
            return NotImplemented

        return Solution(
            # Combine files - create new list to avoid modifying originals
            files=[source_file.file_path for source_file in self._files + other._files],
            # Combine include_dirs - deduplicate by converting to set and back
            include_dirs=list(set(self._include_dirs + other._include_dirs)),
        )

    def patch_solution(self, patch: str) -> None:
        """Patches the solution using the patch command."""
        if not self.verify_solution_integrity():
            raise SolutionIntegrityError(self.files)

        # Save as temp files
        with NamedTemporaryFile(mode="w", delete=False) as patch_file:
            patch_file.write(patch)
            patch_file.flush()

            cmd = [
                "patch",
                "-d",  # Change to working dir first
                str(self.working_dir),
                "-i",  # Flag to specify patch file
                str(Path(patch_file.name).absolute()),
            ]

            process: CompletedProcess = run(
                cmd,
                stdout=PIPE,
                stderr=STDOUT,
                check=False,
            )

            # patch's exit status is 0 if all hunks are applied successfully,
            # 1 if some  hunks  cannot  be applied  or there were merge
            # conflicts, and 2 if there is more serious trouble.
            match process.returncode:
                case 1:
                    print_horizontal_line(get_log_level(0))
                    print(
                        f"The patch:\n\n{patch}\n\nThe diff output:\n\n"
                        + process.stdout.decode("utf-8")
                    )
                    raise ValueError(
                        f"Patch failed for some files in solution {self.working_dir}"
                    )
                case 2:
                    print_horizontal_line(get_log_level(0))
                    print(
                        f"The patch:\n\n{patch}\n\nThe diff output:\n\n"
                        + process.stdout.decode("utf-8")
                    )
                    raise ValueError(
                        f"Patch failed SERIOUSLY in solution {self.working_dir}"
                    )



# Author: Yiannis Charalambous 2023

"""
  This program is designed to create and animate a simple loading animation.
  Original source: https://gist.github.com/rudrathegreat/b11daed176c8119dcedbb6b06c953590
"""

import sys

from time import sleep
from itertools import cycle
from threading import Thread
from types import TracebackType
from typing import Any, AnyStr, BinaryIO, TextIO
from typing_extensions import override
from blessed import Terminal


class BaseLoadingWidget:
    """Base loading widget, will not display any information."""

    @property
    def is_running(self) -> bool:
        """Is the loading animation playing?"""
        return False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _ = exc_type
        _ = exc_val
        _ = exc_tb
        self.stop()

    def __call__(self, text: str | None = None):
        _ = text
        return self

    def start(self, text: str = "") -> None:
        """Starts the animation."""
        _ = text

    def stop(self) -> None:
        """Stops the animation."""


class LoggingWidget(BaseLoadingWidget):
    """Simply prints the messages on the screen."""

    @override
    def start(self, text: str = "") -> None:
        if text:
            print(text)


class _StdoutInterceptor(TextIO):
    """Intercepts writes to sys.stdout so that printing clears and redraws the widget."""

    def __init__(
        self,
        original: TextIO,
        widget: BaseLoadingWidget,
    ) -> None:
        self.original: TextIO = original
        self.widget: BaseLoadingWidget = widget

    @override
    def write(self, text: AnyStr) -> int:
        """Clears the loading widget before writing."""
        # Clear the widget line before writing new text.
        was_running: bool = self.widget.is_running
        if was_running:
            self.widget.stop()

        # Print the text
        result: int = self.original.write(str(text))

        # After printing, redraw the widget if it's still active.
        if was_running:
            self.widget.start()

        return result

    @override
    def __enter__(self) -> TextIO:
        return self.original.__enter__()

    @override
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
        /,
    ) -> None:
        return self.original.__exit__(exc_type, exc_val, exc_tb)

    @property
    @override
    def buffer(self) -> BinaryIO:
        return self.original.buffer

    @property
    @override
    def errors(self) -> str | None:
        return self.original.errors

    @property
    @override
    def line_buffering(self) -> int:
        return self.original.line_buffering

    @property
    @override
    def newlines(self) -> Any:
        return self.original.newlines

    @property
    @override
    def encoding(self) -> str:
        return self.original.encoding

    @override
    def flush(self) -> None:
        """Calls flush on the original stream."""
        return self.original.flush()


loading_widget_anim_1: list[str] = [
    "[          ]",
    "[          ]",
    "[==        ]",
    "[===       ]",
    "[====      ]",
    "[=====     ]",
    "[======    ]",
    "[=======   ]",
    "[========  ]",
    "[========= ]",
    "[==========]",
    "[ =========]",
    "[  ========]",
    "[   =======]",
    "[    ======]",
    "[     =====]",
    "[      ====]",
    "[       ===]",
    "[        ==]",
    "[         =]",
]

loading_widget_anim_2: list[str] = [
    "[ Loading ]",
    "[ Loading ]",
    "[ Loading ]",
    "[ Loading ]",
    "[ Loading ]",
    "[=Loading ]",
    "[==oading ]",
    "[===ading ]",
    "[====ding ]",
    "[=====ing ]",
    "[======ng ]",
    "[=======g ]",
    "[======== ]",
    "[=========]",
    "[ ========]",
    "[ L=======]",
    "[ Lo======]",
    "[ Loa=====]",
    "[ Load====]",
    "[ Loadi===]",
    "[ Loadin==]",
    "[ Loading=]",
    "[ Loading ]",
    "[ Loading ]",
    "[ Loading ]",
    "[ Loading ]",
    "[ Loading ]",
]

loading_widget_anim_3: list[str] = [
    " [=     ]",
    " [ =    ]",
    " [  =   ]",
    " [   =  ]",
    " [    = ]",
    " [     =]",
    " [    = ]",
    " [   =  ]",
    " [  =   ]",
    " [ =    ]",
]

loading_widget_anim_4: list[str] = [
    "     <>",
    "    <==>",
    "   <====>",
    "  <======>",
    " <========>",
    "<==========>",
]

loading_widget_anim_5: list[str] = [
    "     <>",
    "    <==>",
    "   <====>",
    "  <======>",
    " <========>",
    "<==========>",
    " <========>",
    "  <======>",
    "   <====>",
    "    <==>",
]


class LoadingWidget(BaseLoadingWidget):
    """Loading widget that can display an animation along with some text.

    Args:
        anim_speed: Delay in seconds between each frame.
        ephemeral: False to print the text on screen before new text is printed,
            then stop the animation. True to always play the loading animation
            and show the text at the last line of the terminal and new text will
            be shown above.
    """

    def __init__(
        self,
        anim_speed: float = 0.1,
        ephemeral: bool = True,
        animation: list[str] | None = None,
    ) -> None:
        super().__init__()
        self.anim_speed: float = anim_speed
        self.animation: list[str] = animation if animation else ["|", "/", "-", "\\"]
        self.ephemeral: bool = ephemeral
        self.loading_text: str
        self._running: bool = False
        self._terminal: Terminal = Terminal()

        # Last drawn text, is used for _clear_draw
        self._last_text: str = ""
        # Used to track non-ephemeral printing, this variable means that the
        # loading text is already printed.
        self._non_ephemeral_print: bool = False

        # Multi-threading
        self.thread: Thread | None = None
        self._interceptor: _StdoutInterceptor = _StdoutInterceptor(sys.stdout, self)

        # Find the largest frame in the animatioon
        self.anim_clear_length: int = max(len(frame) for frame in self.animation)

    @property
    @override
    def is_running(self) -> bool:
        return self._running

    def __call__(self, text: str | None = None):
        """Allows you to set the text in a with statement easily."""
        if text:
            self.loading_text = text
        return self

    def _clear_draw(self) -> None:
        """Clear the drawn text."""
        self._interceptor.original.write(
            " " * len(self._last_text) + self._terminal.move_left(len(self._last_text)),
        )

    def _animate(self) -> None:
        """We use the original stdout stream because if we call stdout in here
        we will go into the StdoutInterceptor, if that happens during
        thread.join(), then stop() will be called in the write() of interceptor
        and then that will cause a crash because a thread can't call join on
        itself."""
        self._last_text = ""
        for c in cycle(self.animation):
            if not self._running:
                break
            # Calculate how much extra space to clear after anim frame is shown.
            extra_space_clear: int = self.anim_clear_length - len(c)
            self._last_text = f"{self.loading_text} {c}" + " " * extra_space_clear
            # Write + Reset cursor
            self._interceptor.original.write(self._last_text)
            self._interceptor.flush()
            sleep(self.anim_speed)
            self._interceptor.original.write(
                self._terminal.move_left(len(self._last_text))
            )

    @override
    def start(self, text: str = "Please Wait") -> None:
        # Capture stdout
        sys.stdout = self._interceptor
        self._running = True
        self.loading_text = text
        self.thread = Thread(target=self._animate)
        self.thread.daemon = True
        self.thread.start()

    @override
    def stop(self) -> None:
        # Release stdout
        sys.stdout = self._interceptor.original
        self._running = False
        # Block until end.
        if self.thread:
            self.thread.join()

        self._clear_draw()

        if not self.ephemeral and not self._non_ephemeral_print:
            print(self.loading_text)
            self._non_ephemeral_print = True

        self._interceptor.original.flush()

    @override
    def __enter__(self):
        self._non_ephemeral_print = False
        return super().__enter__()


if __name__ == "__main__":
    LoadingWidget(animation=loading_widget_anim_2, ephemeral=True).start()
    while True:
        pass



# Author: Yiannis Charalambous

"""Contains the base class for chat command results."""

from abc import ABC, abstractmethod


class CommandResult(ABC):
    """Base class for the return result of chat commands."""

    @property
    @abstractmethod
    def successful(self) -> bool:
        """Returns true if the execution of the command was successful. False if
        otherwise. If false, then it is up to the command to provide a method to
        access the reason."""
        raise NotImplementedError()

    def __str__(self) -> str:
        return "Command returned " + (
            "successful" if self.successful else "unsuccessful"
        )
